"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from itertools import product

import numpy as np
import networkx as nx
from sympy.physics.quantum import Ket, OrthogonalKet

from qrisp import (QuantumFloat, QuantumBool, QuantumArray, mcz, cx, h, ry, swap, QFT, 
                   auto_uncompute, invert, control, IterationEnvironment, bin_rep, 
                   cyclic_shift, multi_measurement, increment)

"""
As specified in the paper (https://arxiv.org/abs/1509.02374), the key challenge
in implementing the quantum backtracking algorithm is the realization of the operators
R_A and R_B. These operators consists of the direct some of diffuser operators D_x,
where x is an arbitrary node of the backtracking graph.
R_A and R_B are defined as the direct sum of these

R_A = [direct sum] D_x  [summed over all nodes x with even depth]
R_B = |r><r| [direct sum] D_x [summed over all nodes x with even depth]

Or in words: Each node x together with it's children {y : x->y}
defines a subspace H_x = span(|x>, {|y>, x->y}) on which the operator D_x is acting
in-place.

The definition of the D_x operators consists of multiple conditions:
    
    1. If x is an "accept" node, D_x = 1 (ie. the identity)
    2. If x is the root D_x = 1 - 2|psi_r><psi_r|
        Where |psi_r> = (1+d_r*n)**-0.5 * (|r> + n**0.5 * [sum] |y>)
        Where d_r is  the degree of the root, n is the maximum depth of the tree
        and the sum iterates over all children of r
    3. Otherwise: D_x = 1 - 2 |psi_x><psi_x|
        Where |psi_x> = (d_x)**-0.5 (|x> + [sum] |y>)
        Where d_x is the degree of x and the sum iterates over all children of x

To implement this operator, we will rewrite it a bit:
    
D_x = 1 - (1+(-1)**accept(x)) * |psi_r><psi_r|
    
The next step is to assume an operator U_x, which prepares |psi_x> from x

|psi_x> = U_x |x>

We can then write

D_x = D_x = 1 - (1+(-1)**accept(x)) * |psi_r><psi_r|
    = U_x (1 - (1+(-1)**accept(x))*|x><x|) U_x^(-1)

If we pick an encoding, where each node state |x> is a computational basis state,
the center bracket (1 - (1+(-1)**accept(x))*|x><x|) can be realized as a 
multi-controlled Z-gate on the bitstring of |x>, which is also controlled on the 
result of accept(x).

The operator U_x is implemented as the function psi_prep below.

The operator D_x is implemented as the method qstep_diffuser of the
QuantumBacktrackingTree class.

This brings us to another important point in the implementation of the algorithm:
The encoding of the node states.

In principle, the paper makes no statement how such an encoding could be realized.
The approach we took here is the following:
    
A node state |x> is specified by:
    1. A QuantumFloat l, which specifies the distance of x from the root
    2. A QuantumArray branch_qa, which specifies the path to take to reach x.

A few things have to be said about this encoding:
    
The initial state of branch_qa is |0>|0>|0>...

Any state where there is a non-zero state at an index higher than l is considered
"non-algorithmic", ie. the state does not represent a node. An example could be

|0>|0>|1>...|l = 1>

Another important point: As it turns out a much more efficient implementation is 
possible if branch_qa holds the path from the node to the root instead of root-to-node.
For an example of this can be found in the docstring of QuantumBacktrackingTree.

"""



class QuantumBacktrackingTree:
    """
    This class describes the central data structure to run backtracking algorithms in
    a quantum setting. `Backtracking algorithms <https://en.wikipedia.org/wiki/Backtracking>`_ 
    are a very general class of algorithms which cover many problems of combinatorial 
    optimization such as 3-SAT or TSP.
    
    Backtracking algorithms can be put into a very general form. Given is a maximum 
    recursion depth, two functions called ``accept``/``reject`` and the set of 
    possible assignments for an iterable x.
        
    ::
        
        from problem import accept, reject, max_depth, assignments
        
        def backtracking(x):
            
            if accept(x):
                return x
            
            if reject(x) or len(x) == max_depth:
                return None
            
            for j in assigments:
                y = list(x)
                y.append(j)
                res = backtracking(y)
                if res is not None:
                    return res
                
    
    The power of these algorithms lies in the fact that they can quickly discard
    large parts of the potential solution space by using the reject function to
    cancel the recursion. Compared to an unstructured search, where only the 
    accept function is available, this can significantly cut the required resources.
            
    The quantum algorithm for solving these problems has been 
    `proposed by Ashley Montanaro <https://arxiv.org/abs/1509.02374>`_ and yields
    a 1 to 1 correspondence between an arbitrary classical backtracking algorithm
    and it's quantum equivalent. The quantum version achieves a quadratic speed up
    over the classical one.
    
    The algorithm is based on performing quantum phase estimation on a quantum walk
    operator, which traverses the backtracking tree. The core algorithm returns 
    "Node exists" if the 0 component of the quantum phase estimation result 
    has a higher probability then 3/8 = 0.375.
    
    Similar to the classical version, for the Qrisp implementation of this quantum 
    algorithm, a backtracking problem is specified by a maximum recursion depth
    and two functions, each returning a :ref:`QuantumBool` respectively:
    
    **accept**: Is the function that returns True, if called on a node, satisfying the 
    specifications.
    
    **reject**: Is the function that returns True, if called on a node, representing a 
    branch that should no longer be considered.
    
    Furthermore required is a :ref:`QuantumVariable` that specifies the branches
    that can be taken by the algorithm at each node.
    
    .. note::
        
        Many implementations of backtracking also include the possibility for
        deciding which entries of x to assign based on some user provided heuristic.
        The quantum version also supports this feature, however it is not yet
        implemented in Qrisp.

    Parameters
    ----------

    max_depth : integer
        The depth of the backtracking tree.
    branch_qv : QuantumVariable
        A QuantumVariable representing the possible branches of each node.
    accept : function
        A function taking an instance of QuantumBacktrackingTree and returning 
        a QuantumBool, which is True, if called on a satisfying node.
    reject : function
        A function taking an instance of QuantumBacktrackingTree and returning 
        a QuantumBool, which is True, if a called on a node where the corresponding
        branch should no longer be investigated.
        
    
    Attributes
    ----------
    
    h : :ref:`QuantumFloat`
        A QuantumFloat representing the height of the represented node. The root 
        has h = max_depth, it's children have h = max_depth-1 etc.
    branch_qa : :ref:`QuantumArray`
        A QuantumArray representing the path from the root to the current node.
        The qtype of this QuantumArray is what is been provided as branch_qv.
        Note that for implementation efficiency reasons, the path is reversed.
        In a binary tree of depth 4, the node which is reached after taking the
        branches 100 is therefore represented by h = 1, branch_qa = [0,0,1,0].
        States that have a non-zero value at entries indexed bigger than max_depth-h, 
        are considered non-algorithmic and will never be visited 
        (eg. h=3, branch_qa = [1,1,1,1]).
    qs : :ref:`QuantumSession`
        The QuantumSession of the backtracking tree.
    max_depth : int
        An integer specifying the maximum depth of each node.
    
    
    Examples
    --------
    
    We recommend skipping to the "Attributes" section to gain an understanding
    on how the backtracking nodes are encoded into QuantumVariables.
    
    **Checking for the existence of a solution**
    
    Even though the backtracking problems primary purpose is to __find__ a solution,
    at the core, Montanaros algorithm only determines solution existence. This can 
    however still be leveraged into a solution finding algorithm.
    
    To demonstrate the solution existence functionality, we search the binary 
    tree that consists only of nodes with alternating branching.
    We accept if we find the node [1,0,0] (doesn't exist in this tree).
    
    For this we first set up the reject condition. 
    
    ::
        
        from qrisp import *
        
        @auto_uncompute    
        def reject(tree):
            exclude_init = (tree.h < tree.max_depth -1 )
            alternation_condition = (tree.branch_qa[0] == tree.branch_qa[1])
            
            return exclude_init & alternation_condition
        
    Note that the root and it's children are never rejected, or else the algorithm would
    cancel immediately. For the encoding of the nodes please refer to the attributes
    section of this page.
    
    The third condition makes sure that each node where the last branching 
    (tree.branch_qa[0]) is equal to the branching before (tree.branch_qa[1])
    is rejected.
    
    We now implement the accept condition:
    
    ::
        
        @auto_uncompute    
        def accept(tree):
            height_condition = (tree.h == tree.max_depth - 3)
            path_condition = (tree.branch_qa[0] == 0)
            path_condition = path_condition & (tree.branch_qa[1] == 0)
            path_condition = path_condition & (tree.branch_qa[2] == 1)
            
            return height_condition & path_condition
        
    
    Subsequently we set up the class instance:
    
    ::        
        
        from qrisp.quantum_backtracking import QuantumBacktrackingTree
        
        tree = QuantumBacktrackingTree(max_depth = 3,
                                       branch_qv = QuantumFloat(1),
                                       accept = accept,
                                       reject = reject)
        
        tree.init_node([])
        
    We can evaluate the statevector:
        
    >>> tree.statevector()
    1.0*|[]>
    
    The ``[]`` indicates that this is the root state. If the tree was in the state
    of a child of the root (say the one connected to the 1 branch) it would be ``[1]``.
    
    Note that the ``statevector`` method decodes the QuantumVariables holding the
    node state for convenient readibility. If you want to see the encoded variables
    you can take a look at the :ref:`QuantumSession` s :meth:`statevector method<qrisp.QuantumSession.statevector>`:
        
    >>> tree.qs.statevector()
    |0>**4*|3>
    
    We can also visualize the statevector of the tree:
        
    >>> import matplotlib.pyplot as plt
    >>> tree.visualize_statevector()
    >>> plt.show()
    
    .. image:: ./root_state_plot.png
        :width: 200
        :alt: Root statevector plot
        :align: left
    
    |
    |
    |
    |
    |
    
    And finally evaluate the algorithm:
        
    ::
        
        qpe_res = tree.estimate_phase(precision = 4)
    
    Perform a measurement
    
    >>> mes_res = qpe_res.get_measurement()
    >>> mes_res[0]
    0.0253
    
    The 0 component has only 0.0253% probability of appearing, therefore we can conclude,
    that in the specified tree no such node exists.
    
    We now perform the same process but with a trivial reject function:
        
    ::
        
        def reject(tree):
            return QuantumBool()
        
        tree = QuantumBacktrackingTree(max_depth = 3,
                                       branch_qv = QuantumFloat(1),
                                       accept = accept,
                                       reject = reject)
        
        tree.init_node([])
        
        qpe_res = tree.estimate_phase(precision = 4)
        
        
    >>> mes_res = qpe_res.get_measurement()
    >>> mes_res[0]
    0.5027
    
    We see a probability of more than 50%, implying a solution exists in
    this tree.
    
    **Finding a solution**

    Montanaros approach to determine a solution is to classically traverse the tree,     
    by always picking the child node where the quantum algorithm returns "Node exists".
    Finding a solution can therefore be considered a hybrid algorithm.
    
    The Qrisp implementation abuses a little shortcut: After having measured the
    result of the quantum phase estimation, the $\ket{0}$ state is entangled with
    the $\ket{\phi}$ state. This state contains information about the path to
    an accepted node. Therefore by measuring also the tree variables, we
    can make an informed decision which path to take. This implies, that in the 
    limit of infinite shots, only a single query is needed!
    
    To demonstrate, we search for the node [1,1,1] with a trivial reject function.
    
    ::
        
        @auto_uncompute    
        def accept(tree):
            height_condition = (tree.h == tree.max_depth - 3)
            path_condition = QuantumBool()
            mcx(tree.branch_qa[:3], path_condition)
            
            return height_condition & path_condition
        
        def reject(tree):
            return QuantumBool()
        
    
    Set up the QuantumBacktrackingTree instance:
        
        
    >>> max_depth = 4
    >>> tree = QuantumBacktrackingTree(max_depth,
                                       branch_qv = QuantumFloat(1),
                                       accept = accept,
                                       reject = reject)
    
    And call the solution finding algorithm
    
    >>> tree.find_solution(precision = 5)
    [1, 1, 1]
    
        
    """
    
    def __init__(self, max_depth, branch_qv, accept, reject):
        
        self.max_depth = max_depth
        
        self.degree = 2**branch_qv.size
        
        self.branch_qa = QuantumArray(qtype = branch_qv, shape = max_depth)
        
        h_size = int(np.ceil(np.log2(max_depth+2)))

        self.branch_workspace = branch_qv.duplicate(qs = self.branch_qa.qs, name = "branch_workspace*")

        self.h = QuantumFloat(h_size, name = "h*", qs = self.branch_qa.qs)
        
        self.qs = self.h.qs
        
        self.accept_function = accept
        self.reject_function = reject
    
    def accept(self):
        return self.accept_function(self)
    
    def reject(self):
        return self.reject_function(self)

    @auto_uncompute
    def qstep_diffuser(self, even = True, ctrl = []):
        """
        Performs the operators :math:`R_A` or :math:`R_B`. For more information on these operators check `the paper <https://arxiv.org/abs/1509.02374>`_.

        Parameters
        ----------
        even : bool, optional
            If set to ``True`` :math:`R_A` will be performed. Otherwise :math:`R_B` will be performed. The default is True.
        ctrl : List[Qubit], optional
            A list of qubits that allows performant controlling. The default is [].

        Examples
        --------
        
        We set up a QuantumBackTrackingTree and perform the diffuser on a marked node 
        
        ::
            
            from qrisp import auto_uncompute, QuantumBool, QuantumFloat
            from qrisp.quantum_backtracking import QuantumBacktrackingTree
            
            @auto_uncompute    
            def reject(tree):
                return QuantumBool()
            
            @auto_uncompute    
            def accept(tree):
                return (tree.h == 1)
            
            
            tree = QuantumBacktrackingTree(3, QuantumFloat(1, name = "branch_qf*"), accept, reject)
            
            tree.init_node([1,1]) 
        
        >>> print(tree.qs.statevector())
        |1>**2*|0>*|2>*|0>
        >>> tree.qstep_diffuser()
        >>> print(tree.qs.statevector())
        |1>**2*|0>*|2>*|0>
        
        We see that the node (as expected) is invariant under :math:`R_A`.
        """
        
        
        #This function performs the operation
        # D_x = U_x (1 - (1+(-1)**accept(x))*|x><x|) U_x^(-1)
        #For more information, check the beginning of this file
        
        
        #Perform U_x^(-1)
        with invert():
            psi_prep(self, even = even)

        #We now perform the operation
        # 1 - (1+(-1)**accept(x))*|x><x|
        #by executing an appropriate mcz gate
        
        #Two additional details to consider:
        #   1. We allow for additional control qubits. This is important because 
        #   this operator will undergo phase estimation and we want to prevent 
        #   performing automatic synthesis of the controlled operation, 
        #   as this would imply controlling EVERY gate. Instead we can just
        #   control the mcz gate.
        #   2. If reject(x) = True, this node has no children, implying
        #   D_x = 1 - 2*|x><x|
        #   We realize this behavior by adding an additional control on the value
        #   of the reject function to the mcz gate, such that if reject(x) = True
        #   the operator U_x mcz U_x^(-1) = 1
        #   We then realize D_x = 1 - 2*|x><x| with another mcz gate.
        #   This has the disadvantage, that the reject function needs to be 
        #   evaluated twice. If you find a better solution, that requires only a
        #   single evaluation feel free to message us
        #   raphael.seidel@fokus.fraunhofer.de
        

        #This list will hold the qubit for the first mcz gate.
        mcz_list = []
        
        #D_x operates on the space span(|x>, {|y>, x->y})
        #In order to make sure our mcz gate only marks |x>, we can use the
        #oddity of l, because if l(x) is odd, then l(y) is not.
        oddity_qb = self.h.significant(0)
        mcz_list.append(oddity_qb)
        
        #Prepare control state specificator
        if even:
            #If we are in the even case, we want to tag only the states with
            #even l, so we want oddity_qb to be in state 0
            ctrl_state = "0"
        else:
            ctrl_state = "1"
        
        
        #It turns out U_x^(-1) also produces non-algorithmic states, but we
        #only want to tag |x>. Luckily, the non algorithmic states can be identified
        #by checking wether the branch_workspace QuantumVariable is 0. For more
        #details why this is the case, examine the role of this QuantumVariable
        #in the function psi_prep.
        # is_valid = (self.branch_workspace == 0)
        # mcz_list.append(is_valid)
        mcz_list.append(list(self.branch_workspace))
        ctrl_state += "0"*len(self.branch_workspace)
        
        #Determine accept value
        accept_value = self.accept()
        accept_value.flip()
        mcz_list.append(accept_value)
        ctrl_state += "1"
        
        #Determine reject value
        reject_value = self.reject()
        reject_value.flip()
        mcz_list.append(reject_value)
        ctrl_state += "1"
        
        
    	#Add additional control qubits        
        mcz_list += ctrl
        ctrl_state += "1"*len(ctrl)
        
        #Perform mcz gate
        mcz(mcz_list, ctrl_state = ctrl_state)
        
        #Perform U_x
        psi_prep(self, even = even)
        
        
        #Perpare the qubits for the second mcz gate
        mcz_list = [oddity_qb]
        
        #We need to evaluate reject again, because psi_prep 
        #changed the state of the tree
        reject_value = self.reject()
        mcz_list.append(reject_value)
        
        #Add additional control qubits
        mcz_list += ctrl
        
        #Specify the control state for the second mcz
        if even:
            ctrl_state = "0"
        else:
            ctrl_state = "1"
        
        ctrl_state += (len(mcz_list)-1)*"1"
        
        #Perform the second mcz
        mcz(mcz_list, ctrl_state = ctrl_state)
        
        
        
    def quantum_step(self, ctrl=[]):
        """
        Performs the operator :math:`R_B R_A`. For more information check the :meth:`diffuser methode <qrisp.quantum_backtracking.QuantumBacktrackingTree.qstep_diffuser>`.

        Parameters
        ----------
        ctrl : List[Qubit], optional
            A list of qubits, the step operator should be controlled on. The default is [].
        """
        
        self.qstep_diffuser(even = self.max_depth%2, ctrl=ctrl)
        self.qstep_diffuser(even = not self.max_depth%2, ctrl=ctrl)
        
    def estimate_phase(self, precision):
        r"""
        Performs :meth:`quantum phase estimation <qrisp.QPE>` on the :meth:`quantum step operator <qrisp.quantum_backtracking.QuantumBacktrackingTree.quantum_step>`.
        
        If executed with sufficient precision, the phase estimation will yield a QuantumFloat, where the probability of the 0 component indicates the presence of a node where the ``accept`` function yielded ``True``.
        
        If the probability is higher than 3/8 :math:`\Rightarrow` A solution exists.
        
        If the probability is less than 1/4 :math:`\Rightarrow` No solution exists.
        
        Otherwise :math:`\Rightarrow` Increase precision.
        
        In general, the required precision is proportional to
        
        .. math::
            
            \frac{\text{log}_2(Tn)}{2} + \beta
            
        Where :math:`T` is the amount of nodes, that would be visited by a classical algorithm, :math:`n` is the maximum depth and :math:`\beta` is a universal constant.
        

        Parameters
        ----------
        precision : int
            The precision to perform the quantum phase estimation with.

        Returns
        -------
        qpe_res : :ref:`QuantumFloat`
            The QuantumFloat containing the result of the phase estimation.

        """
        
        qpe_res = QuantumFloat(precision, -precision, qs = self.qs)

        h(qpe_res)

        for i in range(qpe_res.size):
            with IterationEnvironment(self.qs, 2**i, precompile = True):
                self.quantum_step(ctrl = [qpe_res[i]])
                
        QFT(qpe_res, inv=True)
        
        return qpe_res

    
    def init_phi(self, path):
        r"""
        Initializes the normalized version of the state :math:`\ket{\phi}`.
        
        .. math::
            
            \ket{\phi} = \sqrt{n}\ket{r} + \sum_{x \neq r, \\ x \rightsquigarrow x_0} (-1)^{l(x)} \ket{x}

        Where :math:`x \rightsquigarrow x_0` means that :math:`x` is on the path from :math:`r` to :math:`x_0` (including :math:`x_0`).
        
        If :math:`x_0` is a marked node, this state is invariant under the quantum step operator.
        
        Parameters
        ----------
        path : List
            The list of branches specifying the path to :math:`x_0`.

        Examples
        --------
        
        We set up a backtracking tree of depth 3, where the marked element is the 111 node.
        
        ::
            
            from qrisp import auto_uncompute, QuantumBool, QuantumFloat
            from qrisp.quantum_backtracking import QuantumBacktrackingTree
            
            @auto_uncompute    
            def reject(tree):
                return QuantumBool()
            
            @auto_uncompute    
            def accept(tree):
                return (tree.branch_qa[0] == 1) & (tree.branch_qa[1] == 1) & (tree.branch_qa[2] == 1)
            
            
            tree = QuantumBacktrackingTree(3, QuantumFloat(1, name = "branch_qf*"), accept, reject)
        
        Initialize :math:`\ket{\phi}` and evaluate the statevector:
        
        >>> tree.init_phi([1,1,1])
        >>> print(tree.qs.statevector())
        (1.0*sqrt(2)*|0>**5 - 0.816496014595032*|1>*|0>**2*|1>*|0> + 0.816496014595032*|1>**2*|0>*|2>*|0> - 0.816496014595032*|1>**3*|3>*|0>)/2
        
        Perform the quantum step and evaluate the statevector again:
        
        >>> tree.quantum_step()
        >>> print(tree.qs.statevector())
        (1.0*sqrt(2)*|0>**5 - 0.816489994525909*|1>*|0>**2*|1>*|0> + 0.816489994525909*|1>**2*|0>*|2>*|0> - 0.816489994525909*|1>**3*|3>*|0>)/2
        
        We see that the node (as expected) is invariant under the quantum step operator.
        """
        
        h_state = {}
        h_state[self.max_depth] = (self.max_depth)**0.5

        for i in range(1, len(path)+1):
            h_state[self.max_depth - i] = (-1)**(i)

        self.h[:] = h_state

        for i in range(1, len(path)+1):
            with self.h == self.max_depth - i:
                for j in range(i):
                    self.branch_qa[j].encode(path[::-1][j], permit_dirtyness = True)
                    
    def init_node(self, path):
        """
        Initializes the state of a given node.

        Parameters
        ----------
        path : List
            List of the branch labels indicating the path from the root to the node.

        Examples
        --------
        
        We initialize a backtracking tree in the 101 node.
        
        ::
            
            from qrisp import auto_uncompute, QuantumBool, QuantumFloat
            from qrisp.quantum_backtracking import QuantumBacktrackingTree
            
            @auto_uncompute    
            def reject(tree):
                return QuantumBool()
            
            @auto_uncompute    
            def accept(tree):
                return QuantumBool()
            
            tree = QuantumBacktrackingTree(3, QuantumFloat(1, name = "branch_qf*"), accept, reject)
            
            tree.init_node([1,0,1])

        """
        
        self.h[:] = self.max_depth - len(path)
        if len(path):
            self.branch_qa[:len(path)] = path[::-1]
    
    def subtree(self, new_root):
        """
        Returns the subtree of a given node.

        Parameters
        ----------
        new_root : list
            The path from the root of self to the root of the subtree.

        Returns
        -------
        QuantumBacktrackingTree
            The subtree starting at the specified root.
        
        Examples
        --------
        
        We initiate a QuantumBacktrackingTree with trivial reject
        function and create a subtree starting at an accepted node-
        
        ::
            
            from qrisp import auto_uncompute, QuantumBool
            from qrisp.quantum_backtracking import QuantumBacktrackingTree
            
            @auto_uncompute    
            def accept(tree):
                height_cond = (tree.h == 2)
                return height_cond

            @auto_uncompute    
            def reject(tree):
                return QuantumBool()
            
        
        Create and initiate the parent tree.
        
        >>> depth = 5
        >>> tree = QuantumBacktrackingTree(depth, QuantumFloat(1, name = "branch_qf*"), accept, reject)
        >>> tree.init_node([])
        >>> print(accept(tree))
        {False: 1.0}
        
        We now create the subtree, where the new root has height two, ie. the accept
        function returns ``True``.
        
        >>> subtree = tree.subtree([0,1,0])
        >>> subtree.init_node([])
        >>> print(accept(subtree))
        {True: 1.0}

        """
        return Subtree(self, new_root)
    
    def copy(self):
        """
        Returns a copy of self. Copy means a QuantumBacktrackingTree with identical
        depth, accept/reject functions etc. but with freshly allocated QuantumVariables.

        Returns
        -------
        QuantumBacktrackingTree
            Another instance with the same depth/accept/reject etc.

        """
        return Subtree(self, [])
    
    def find_solution(self, precision, cl_accept = None, measurement_kwargs = {}):
        """
        Determines a path to a solution.

        Parameters
        ----------
        precision : integer
            The precision to perform the quantum phase estimation(s) with.
        cl_accept : function, optional
            A classical version of the accept function of self. Needs to 
            receive a list to indicate a path and returns a bool wether the 
            node is accepted. By default, the accept function of self will be
            evaluated on a simulator.
        measurement_kwargs : dictionary
            A dictionary to give keyword arguments that specify how measurements
            are evaluated. The default is {}.

        Returns
        -------
        List
            A list indicating the path to a node where the accept function
            returns True.

        Examples
        --------
        
        We create a accept function that marks the node [0,1] and a trivial 
        reject function. 
        
        ::
            
            from qrisp import auto_uncompute, QuantumBool, QuantumFloat, mcx
            from qrisp.quantum_backtracking import QuantumBacktrackingTree
            
            @auto_uncompute    
            def accept(tree):
                height_cond = (tree.h == 1) # The [0,1] node has height 1
                path_cond = QuantumBool()
                mcx(list(tree.branch_qa)[:-1], path_cond, ctrl_state="10")
                return path_cond & height_cond

            @auto_uncompute    
            def reject(tree):
                return QuantumBool()


        Create backtracking tree object:

        >>> depth = 3
        >>> tree = QuantumBacktrackingTree(depth, QuantumFloat(1, name = "branch_qf*"), accept, reject)
        
        Find solution
        
        >>> res = tree.find_solution(4)
        >>> print(res)
        [0, 1]
        
        """
        
        return find_solution(self, precision, cl_accept, measurement_kwargs=measurement_kwargs)

        
    def path_decoder(self, h, branch_qa):
        """
        Returns the path representation for a given constellation of the
        ``h`` and ``branch_qa`` variables. The path representation is a list
        indicating which branches to take starting from the root.
        This function exists because the encoding of the nodes is hardware 
        efficient but inconvenient for humans to read.

        Parameters
        ----------
        h : integer
            The integer describing the height of the node.
        branch_qa : list
            The list of branches to take to reach the root, starting from the node.

        Returns
        -------
        list
            The list of path variables to take to reach the node, starting 
            from the root.

        Examples
        --------
        
        We create a QuantumBacktrackingTree, initiate a node and retrieve the path.
        
        ::
            
            from qrisp import auto_uncompute, QuantumBool, QuantumFloat, multi_measurement
            from qrisp.quantum_backtracking import QuantumBacktrackingTree

            @auto_uncompute    
            def accept(tree):
                return QuantumBool()
            
            @auto_uncompute    
            def reject(tree):
                return QuantumBool()
            
        >>> depth = 5
        >>> tree = QuantumBacktrackingTree(depth, QuantumFloat(1, name = "branch_qf*"), accept, reject)
        >>> tree.init_node([1,0])
        >>> multi_measurement([tree.h, tree.branch_qa])
        {(3, OutcomeArray([0, 1, 0, 0, 0])): 1.0}
        
        Retrieve the path
        
        >>> tree.path_decoder(3, [0,1,0,0,0])
        [1, 0]

        """
        
        l = self.max_depth - h
        return list(branch_qa[:l][::-1])

        
                    
    def statevector_graph(self, return_root = False):
        r"""
        Returns a NetworkX Graph representing the quantum state of the backtracking tree. 
        The nodes have an ``amplitude`` attribute, indicating the complex amplitude of that node.

        Parameters
        ----------
        return_root : bool, optional
            If set to ``True``, this method will also return the root node. The default is False.

        Returns
        -------
        networkx.DiGraph
            A graph representing the statevector.
            
        Examples
        --------
        
        We initialize a backtracking tree, initialize a :math:`\ket{\phi}` state and retrieve
        the statevector graph.
        
        ::
            
            from qrisp import auto_uncompute, QuantumBool, QuantumFloat
            from qrisp.quantum_backtracking import QuantumBacktrackingTree
            
            @auto_uncompute    
            def reject(tree):
                return QuantumBool()
            
            @auto_uncompute    
            def accept(tree):
                return (tree.branch_qa[0] == 1) & (tree.branch_qa[1] == 1) & (tree.branch_qa[2] == 1)
        
            
        Create backtracking tree and initialize :math:`\ket{\phi}`.
        
        >>> tree = QuantumBacktrackingTree(3, QuantumFloat(1, name = "branch_qf*"), accept, reject)
        >>> tree.init_phi([1,1,1])
        >>> statevector_graph = tree.statevector_graph()
        >>> print(statevector_graph.nodes())
        [QBTNode(path = [], amplitude = (0.7071057-2e-08j)), QBTNode(path = [0], amplitude = 0j), QBTNode(path = [1], amplitude = (-0.40824768-1e-08j)), QBTNode(path = [0, 0], amplitude = 0j), QBTNode(path = [0, 1], amplitude = 0j), QBTNode(path = [1, 0], amplitude = 0j), QBTNode(path = [1, 1], amplitude = (0.4082477+4e-08j)), QBTNode(path = [0, 0, 0], amplitude = 0j), QBTNode(path = [0, 0, 1], amplitude = 0j), QBTNode(path = [0, 1, 0], amplitude = 0j), QBTNode(path = [0, 1, 1], amplitude = 0j), QBTNode(path = [1, 0, 0], amplitude = 0j), QBTNode(path = [1, 0, 1], amplitude = 0j), QBTNode(path = [1, 1, 0], amplitude = 0j), QBTNode(path = [1, 1, 1], amplitude = (-0.4082477+0j))]
        >>> statevector_graph, root = tree.statevector_graph(return_root = True)
        >>> print(root)
        QBTNode(path = [], amplitude = (0.7071057-2e-08j))

        """
        
        from networkx import DiGraph
        
        sv_function = self.qs.statevector("function")
        
        res_graph = DiGraph()
        
        root = QBTNode(self, [])
        
        root.amplitude = sv_function(root.sv_specifier())
        
        res_graph.add_node(root)
        
        last_layer = [root]
        
        
        for i in range(self.max_depth):
            
            next_layer = []
            
            for parent_node in last_layer:
                
                for j in range(2**self.branch_qa[0].size):
                        
                    child_node_path = list(parent_node.path) + [self.branch_qa[0].decoder(j)]

                    child_node = QBTNode(self, child_node_path)
                    
                    child_node.amplitude = sv_function(child_node.sv_specifier())
                    
                    res_graph.add_node(child_node)
                    res_graph.add_edge(parent_node, child_node, label = child_node_path[-1])
                    
                    next_layer.append(child_node)
                    
            last_layer = next_layer
        
        if return_root:
            return res_graph, root
        return res_graph
    
    def visualize_statevector(self, pos = None):
        """
        Visualizes the statevector graph.

        Parameters
        ----------
        pos : dict, optional
            A dictionary indicating the positional layout of the nodes. For more information visit 
            `this page <https://networkx.org/documentation/stable/reference/generated/networkx.drawing.nx_pylab.draw.html#networkx.drawing.nx_pylab.draw>`_. 
            By default is suitable will be generated.

        Examples
        --------
        
        We initialize a backtracking tree and visualize:
            
        ::
            
            from qrisp import auto_uncompute, QuantumBool, QuantumFloat
            from qrisp.quantum_backtracking import QuantumBacktrackingTree
            import matplotlib.pyplot as plt
            
            @auto_uncompute    
            def reject(tree):
                return QuantumBool()
            
            @auto_uncompute    
            def accept(tree):
                return (tree.branch_qa[0] == 1) & (tree.branch_qa[1] == 1) & (tree.branch_qa[2] == 1)
            
            
        >>> tree = QuantumBacktrackingTree(3, QuantumFloat(1, name = "branch_qf*"), accept, reject)
        >>> tree.visualize_statevector()
        >>> plt.show()
        
        .. image:: ./root_state_plot.png
            :width: 200
            :alt: Root statevector plot
            :align: left
        
        >>> tree.init_phi([1,1,1])
        >>> tree.visualize_statevector()
        
        .. image:: ./phi_state_plot.png
            :width: 200
            :alt: Root statevector plot
            :align: left
        

        """
        
        G, root = self.statevector_graph(return_root = True)
        
        def tree_layout(G, node, depth, theta_parent, res_dic = {}):
            
            r = depth + 1
            delta_theta = 2*np.pi/self.degree**(depth+1)
            theta_start = theta_parent - 2*np.pi/self.degree**(depth)/4
            
            children = list(G.neighbors(node))
            
            for i in range(len(children)):
                
                theta = theta_start + i*delta_theta
                
                res_dic[children[i]] = (r*np.sin(theta), r*np.cos(theta))
                
                tree_layout(G, children[i], depth+1, theta, res_dic)
                
            return res_dic
            
        
        pos=tree_layout(G, root, 0, 0)
        pos[root] = (0,0)

        import colorsys
        def complex_to_color(cnumber):
            
            
            angle = np.angle(cnumber)
            radius = np.abs(cnumber)
            
            # Normalize the angle to the range [0, 2*pi)
            angle = (angle +np.pi*5/2)% (2 * np.pi)
        
            # Map the angle to the hue component of the color
            hue = angle / (2*np.pi)
        
            # Map the radius to the saturation and value components of the color
            saturation = 1
            value = 1
        
            # Convert the HSV components to RGB
            hsv_color = (hue, saturation, value)
            rgb_color = colorsys.hsv_to_rgb(*hsv_color)
            
            from scipy.special import expit
            
            intensity = expit((radius-0.2)*10)
            
            
            # Convert the RGB components to hexadecimal format
            hex_color = '#{:02x}{:02x}{:02x}'.format(
                int(rgb_color[0] * 255*intensity),
                int(rgb_color[1] * 255*intensity),
                int(rgb_color[2] * 255*intensity)
            )
        
            return hex_color

        colors = [complex_to_color(node.amplitude) for node in G.nodes()]

        nx.draw(G, pos)
        nx.draw_networkx_nodes(G, pos, node_color = colors)

        edge_labels = dict([((n1, n2), l)
                            for n1, n2, l in G.edges(data = "label")])

        nx.draw_networkx_edge_labels(G, pos=pos, edge_labels=edge_labels)
    
    def statevector(self):
        """
        Returns a SymPy statevector object representing the state of the tree 
        with decoded node labels.

        Returns
        -------
        state : sympy.Expr
            A SymPy quantum state representing the statevector of the tree.

        Examples
        --------
        
        We create a QuantumBacktrackingTree with and investigate the action of the 
        :meth:`quantum step diffuser <qrisp.quantum_backtracking.QuantumBacktrackingTree.qstep_diffuser>`
        
        ::
            
            from qrisp import auto_uncompute, QuantumBool, QuantumFloat
            from qrisp.quantum_backtracking import QuantumBacktrackingTree
            
            @auto_uncompute    
            def accept(tree):
                height_cond = (tree.h == 1) # The [0,1] node has height 1
                path_cond = QuantumBool()
                mcx(list(tree.branch_qa)[:-1], path_cond, ctrl_state="10")
                return path_cond & height_cond

            @auto_uncompute    
            def reject(tree):
                height_cond = (tree.h == 2) # The [1] node has height 2
                path_cond = QuantumBool()
                mcx(list(tree.branch_qa)[0], path_cond, ctrl_state="1")
                return path_cond & height_cond

        
        Create tree and initialize a node where neither accept nor reject are True.
        
        >>> tree = QuantumBacktrackingTree(3, QuantumFloat(1, name = "branch_qf*"), accept, reject)
        >>> tree.init_node([0,0])
        
        Evaluate statevector
        
        >>> print(tree.statevector())
        1.0*|[0, 0]>
        >>> tree.qstep_diffuser(even = False)
        >>> print(tree.statevector())
        -0.666660010814667*|[0, 0, 0]> - 0.666660010814667*|[0, 0, 1]> + 0.333330005407333*|[0, 0]>
        
        We see that the :meth:`quantum step diffuser <qrisp.quantum_backtracking.QuantumBacktrackingTree.qstep_diffuser>`
        moves the state to the children of the [0,0] node (ie. [0,0,0] and [0,0,1]).
        
        We now investigate how it behaves on nodes that are accepted/rejected:
            
        Initiate a new tree
        
        >>> tree = tree.copy()
        >>> tree.init_node([0,1])
        >>> tree.qstep_diffuser(even = False)
        >>> tree.statevector()
        1.0*|[0, 1]>
        
        As expected, the accepted node is invariant.
        
        To investigate the rejected node, we create another copy:
            
        >>> tree = tree.copy()
        >>> tree.init_node([1])
        >>> tree.qstep_diffuser(even = True)
        >>> tree.statevector()
        -0.999989986419678*|[1]>
        
        As expected, the node has eigenvalue -1 (up to numerical precision).
        
        If you are unsure why these statevectors are eigenvector please check 
        `the paper <https://arxiv.org/abs/1509.02374>`_.
        
        """
        
        sv_function = self.qs.statevector("function", decimals = 10)
        
        #Internal qvs are the quantum variables that specify a backtrackingtree node
        internal_qvs = [self.h, self.branch_workspace] + list(self.branch_qa)
        #External qvs are any quantum variables that are also registered in the QuantumSession
        #but don't specify a node
        external_qvs = list(self.qs.qv_list)
        
        #Remove the internal qvs from the external qv_list        
        for qv in internal_qvs:
            for i in range(len(external_qvs)):
                if hash(external_qvs[i]) == hash(qv):
                    external_qvs.pop(i)
                    break
        
        #Get a list of possible labels for each qv
        
        internal_qv_labels = []
        for qv in internal_qvs:
            label_list = []
            for i in range(2**qv.size):
                label_list.append(qv.decoder(i))
            internal_qv_labels.append(label_list)
        
        external_qv_labels = []
        for qv in external_qvs:
            label_list = []
            for i in range(2**qv.size):
                label_list.append(qv.decoder(i))
            external_qv_labels.append(label_list)
        
        #This retrieves a list of all possible constellations of labels
        internal_label_product = list(product(*internal_qv_labels))
        external_label_product = list(product(*external_qv_labels))
        
        #This will be the sympy object that is returned
        res_state = 0

        #Go through all internal label consteallations (ie. node states)        
        for internal_label_const in internal_label_product[::-1]:
            
            #Get the path to that node state
            path = self.path_decoder(internal_label_const[0], internal_label_const[2:])
            
            #Create a label dic for the sv_function
            internal_label_dic = {internal_qvs[i] : internal_label_const[i] for i in range(len(internal_qvs))}
            
            #If there are no external qvs, we can simply call the sv_function
            #with the label dic                        
            if len(external_qvs) == 0:
                
                amplitude = sv_function(internal_label_dic)
                
                if abs(amplitude) < 1E-5:
                    continue
                
                #Add the corresponding ket                
                res_state  += np.round(amplitude, 5) * OrthogonalKet(str(path))
            
            
            #If there are external qvs we do a similar procedure to go through
            #all label constellations
            else:
                
                for external_label_const in external_label_product:
                    
                    
                    #Set up the external label dic
                    external_label_dic = {external_qvs[i] : external_label_const[i] for i in range(len(external_qvs))}
                    
                    #Integrate the internal label dic
                    external_label_dic.update(internal_label_dic)
                    
                    #Retrieve the amplitude
                    amplitude = sv_function(external_label_dic)
                    
                    if abs(amplitude) < 1E-5:
                        continue
                    
                    external_ket_expr = 1
                    #Generate the ket expression for the external qvs                    
                    for label in external_label_const:
                        external_ket_expr *= OrthogonalKet(label)
                    
                    #Add the corresponding state
                    res_state += amplitude * OrthogonalKet(str(path)) * external_ket_expr
            
        return res_state 
        
        
        
"""
This function realizes the operator U_x, which has the property

U_x |x> = |psi_x>

For more details on these objects, check the details of the 
paper (https://arxiv.org/abs/1509.02374)or the beginning of this file.

The general idea to implement this operator are the following two steps:
    
    1. Manipulatre l such that |l> -> 1/N(|l> + c*|l+1>) with suitable N,c \in R
    2. Manipulate branch_qa controlled on l+1 to incorporate the new branches
    
The value of the constant c depends on wether x is the root or a terminal node.

For terminal nodes, c = 0.    
For the root c = n**0.5*d_x**0.5
Otherwise c = d_x**0.5

Where d_x is the degree of the node.

"""             

def psi_prep(x, even = True):
    
    #Determine c
    c = x.degree**0.5
    
    #The step |h> -> 1/N(|h> + c*|h-1>) will be performed by a ry gate
    phase = -np.arctan(c)*2
    root_phase = -np.arctan((c*np.sqrt(x.max_depth)))*2
    
    #We distinguish wether h is odd. This is because if h is odd,
    #the ry gate on the significance 0 qubit acts different
    oddity_qb = x.h.significant(0) 
    
    
    #Distinguish the cases wether psi prepares an even node
    if not even:
        
        #Because the h attribute describes the height of the tree, 
        #the root node can have even or odd h depending on what the maximum
        #depth of the tree is.
        #If the maximum depth is odd, the root node is also odd.
        if x.max_depth%2:
            
            #Perform the ry rotation
            ry(phase, oddity_qb)
            
            #If x is the root (ie. h = x.max_depth), we reverse the previous 
            #ry gate and instead perform the root phase
            with control(x.h[1:], ctrl_state = x.max_depth>>1):
                ry(root_phase - phase, oddity_qb)
            
            #This step initializes the new branch_qa entry.
            with control(oddity_qb, ctrl_method = "gray", ctrl_state = "0"):
                #The QuantumVariable x.branch_workspace will always be in the |0> state
                #at this point of the algorithm, so we swap that |0> state to the first
                #entry of branch_qa and initialize using H-gates to reach uniform superposition.
                cyclic_shift(list(x.branch_qa) + [x.branch_workspace])
                
                h(x.branch_qa[0])
                
        #If the height is even, the h value of the root is also even, therefore
        #the odd case of the psi_prep function does not need to deal with it.
        #Instead it should leave the root invariant, since the state with
        #h = max_depth + 1 is non algorithmic
        else:
            
            #This control environment checks wether h is the root state
            #and only executes it's content if this is not the case
            with control(x.h[1:], ctrl_state = x.max_depth>>1, invert = True):
                
                ry(phase, oddity_qb)
                
                #Initialize branch_qa entry
                with control(oddity_qb, ctrl_method = "gray", ctrl_state = "0"):
                    cyclic_shift(list(x.branch_qa) + [x.branch_workspace])
                    h(x.branch_qa[0])
            
    else:
        
        #For the even case, the oddity qubit will be in the |0> state, so our
        #approach of simply applying an ry will not work anymore as it would
        #perform the operation |h> -> 1/N(|h> + c|h+1>) ie. it would move upward
        
        #To circumvent, we temporarily decrement h such that the oddity qubit is
        #in the |1> state.
        
        increment(x.h, -1)
        #This control environment makes sure that h = 0 nodes 
        #(after decrementation h = -1) are not subjected to an ry gate
        with control(x.h[1:], ctrl_state = -1):
            ry(-phase, oddity_qb)
        
        #Similar to the even case, we distinguish between the cases of the tree having
        #even or odd maximum depth
        if not x.max_depth%2:
            
            #Apply the ry rotation
            ry(phase, oddity_qb)
                
            #If x is the root (ie. h = x.max_depth-1), we reverse the previous 
            #ry gate and instead perform the root phase
            with control(x.h[1:], ctrl_state = (x.max_depth-1)>>1):
                ry(root_phase - phase, oddity_qb)
            
            #Initialize the new branch_qa entry
            with control(oddity_qb, ctrl_state = "0", ctrl_method = "gray"):
                
                cyclic_shift(list(x.branch_qa) + [x.branch_workspace])
                h(x.branch_qa[0])
            
        else:
            
            #Perform a similar logic as for the odd case
            with control(x.h[1:], ctrl_state = (x.max_depth-1)>>1, invert = True):
                ry(phase, oddity_qb)
            
                with control(oddity_qb, ctrl_state = "0", ctrl_method = "gray"):
                    
                    cyclic_shift(list(x.branch_qa) + [x.branch_workspace])
                    h(x.branch_qa[0])
        
        #Reverse the decrementation
        increment(x.h, 1)


class Subtree(QuantumBacktrackingTree):
    
    def __init__(self, parent_tree, root_path):
        
        if len(root_path) > parent_tree.max_depth:
            raise Exception("Tried to initialise subtree with root path longer than maximum depth")
        
        QuantumBacktrackingTree.__init__(self, 
                                         parent_tree.max_depth - len(root_path), 
                                         parent_tree.branch_workspace,
                                         parent_tree.accept_function,
                                         parent_tree.reject_function
                                         )
        
        self.embedding_array = QuantumArray(qtype = self.branch_qa.qtype, shape = len(root_path), qs = self.qs)
        self.branch_qa = np.concatenate((self.branch_qa, self.embedding_array))
        
        self.root_path = root_path
        self.original_tree = parent_tree
    
    def init_node(self, path):
        
        self.h[:] = self.max_depth - len(path)
        
        path = self.root_path + path
        
        if len(path):
            self.branch_qa[:len(path)] = path[::-1]
    
    def init_phi(self, path):
        
        h_state = {}
        h_state[self.max_depth] = (self.max_depth)**0.5

        for i in range(1, len(path)+1):
            h_state[self.max_depth - i] = (-1)**(i)

        self.h[:] = h_state
        
        with self.h == self.max_depth:
            for k in range(len(self.root_path)):
                self.branch_qa[k][:] = self.root_path[::-1][k]

        for i in range(1, len(path)+1):
            with self.h == self.max_depth - i:
                for j in range(i):
                    self.branch_qa[j].encode(path[::-1][j], permit_dirtyness = True)
                    
                for k in range(len(self.root_path)):
                    self.branch_qa[k+i][:] = self.root_path[::-1][k]
                    
    def subtree(self, path):
        return self.original_tree.subtree(self.root_path + path)
        
        
        
def find_solution(tree, precision, cl_accept = None, traversed_nodes = None, measurement_kwargs = {}):
    #The idea of this function is to use the quantum algorithm to check wether
    #a the subtree of a given node contains a solution and then recursively call
    #this function on that subtree.
    
    
    #If there is no classical accept function given, we create a copy of the original
    #tree and evaluate the quantum accept function on that node via the simulator
    if cl_accept is None:
        def cl_accept(path):
            if isinstance(tree, Subtree):
                copied_tree = tree.original_tree.copy()
            else:
                copied_tree = tree.copy()
            copied_tree.init_node(path)
            accept_qbl = copied_tree.accept()
            mes_res = accept_qbl.get_measurement()
            return mes_res == {True : 1}
            
    
    #The first step is to check wether the current root is a solution
    if isinstance(tree, Subtree):
        path = tree.root_path
    else:
        path = []

    if cl_accept(path):
        return []
    elif tree.max_depth == 0:
        return None
    
    #This list keeps track of which nodes have already been checked for solutions
    if traversed_nodes is None:
        traversed_nodes = []

    #Initialize the root node
    tree.init_node([])
    
    #Perform quantum phase estimation
    qpe_res = tree.estimate_phase(precision)
    
    #Retrieve the measurement results
    mes_res = multi_measurement([qpe_res, tree.h, tree.branch_qa], **measurement_kwargs)
    
    #We will first check wether there is a solution    
    #The s variable will contain the probability to measure
    #the qpe_res == 0 branch.
    s = 0
    #This list will contain the possible branches
    new_branches = []
    
    for k, v in mes_res.items():
        #k[0] is the value of qpe_res
        if k[0] == 0:
            #If the measurement result is part of the qpe == 0 branch, add the
            #probability to s
            s += v
            
            new_branches.append(k)
    #If the probability is less then 0.25, there is no solution
    if s <= 0.25:
        return None
    
    #If the probability is between 0.25 and 0.375, the qpe needs more precision
    if s <= 0.375:
        raise Exception("Executed find solution method of quantum backtracking algorithm with insufficient precision")
    
    
    #To find the next node to check we will use a heuristic.
    #After measurement of the 0 branch, the tree is collapsed to a state which 
    #also contains the |phi> state. The phi state is a superposition of nodes 
    #leading to the desired solution.
    #We assume that the outcome with the smallest tree.h value is the state 
    #corresponding to |phi>. This proved to be the case in every situation we tested.
    #If this assumption for whichever reason is not correct, the solution will still
    #be found because of the recursive nature of this algorithm.
    
    #Sort for the value of tree.h
    new_branches.sort(key = lambda x : x[1])
    
    for b in new_branches:
        
        #Get the path to the new node
        new_path = tree.path_decoder(b[1], b[2])
        
        if tuple(path + new_path) in traversed_nodes or len(new_path) == 0:
            continue
        
        #Generate the subtree        
        subtree = tree.subtree(new_path)
        
        #Recursive call
        sub_sol = find_solution(subtree, precision, cl_accept, traversed_nodes)
        
        #If a solution has been found, concatenate the pathes
        if sub_sol is not None:
            solution = path + new_path + sub_sol
            break
        else:
            traversed_nodes.append(tuple(path + new_path))
        
    else:
        raise Exception("Executed find solution method of quantum backtracking algorithm with insufficient precision")

    return solution



class QBTNode:
    
    def __init__(self, tree, path, amplitude = None):
        
        self.h = tree.max_depth - len(path)
        self.path = path
        self.tree = tree
        self.amplitude = amplitude
    
    def __hash__(self):
        return hash(str(self.path))
    
    def sv_specifier(self):
        amplitude_state_specifyer = {self.tree.h : self.tree.max_depth - len(self.path),
                                     self.tree.branch_workspace : self.tree.branch_workspace.decoder(0)}
    
        path = list(self.path)
        if isinstance(self.tree, Subtree):
            path = self.tree.root_path + path
    
        for k in range(len(self.tree.branch_qa)):
            if k < len(path):
                amplitude_state_specifyer[self.tree.branch_qa[k]] = path[::-1][k]
            else:
                amplitude_state_specifyer[self.tree.branch_qa[k]] = 0
        
        
        return amplitude_state_specifyer
    
    def __str__(self):
        return "QBTNode(path = " + str(self.path) + ", amplitude = " + str(self.amplitude) + ")"
    
    def __repr__(self):
        return str(self)



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# Created by ann81984 at 26.07.2022
from qrisp import QuantumVariable, QuantumFloat
import numpy as np
import matplotlib.pyplot as plt
from numpy.linalg import norm


def test_state_preparation():
    qf = QuantumFloat(4, -2, signed=True)

    state_dic = {
        2.75: 1 / 4**0.5,
        -1.5: -1 / 4**0.5,
        2: 1 / 4**0.5,
        3: 1j / 4**0.5,
    }

    qf.init_state(state_dic)

    debugger = qf.qs.statevector("function")

    print(type(debugger))

    print("Amplitude of state 2.75: ", debugger({qf: 2.75}))
    print("Amplitude of state -1.5: ", debugger({qf: -1.5}))
    print("Amplitude of state 3: ", debugger({qf: 3}))

    assert np.abs(debugger({qf: 2.75}) - 0.5) < 1e-5
    assert np.abs(debugger({qf: -1.5}) + 0.5) < 1e-5
    assert np.abs(debugger({qf: 3}) - 0.5j) < 1e-5



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import sys

from qrisp.core.quantum_variable import QuantumVariable
from qrisp.environments import control


class QuantumBool(QuantumVariable):
    """
    QuantumBools are the quantum type, which represents boolean truth values.
    They are the return type of comparison operators like the equality ``==``.

    Apart from their behavior as a QuantumVariable, they can also be treated like
    :ref:`ControlEnvironments <ControlEnvironment>`.

    .. note:
        QuantumBools that are evaluated directly after a ``with`` statement are
        uncomputed automatically upon leaving the ControlEnvironment.

    Examples
    --------

    We create a QuantumBool and set it to uniform superposition

    >>> from qrisp import QuantumBool, h
    >>> q_bool_0 = QuantumBool()
    >>> h(q_bool_0)
    >>> print(q_bool_0)
    {False: 0.5, True: 0.5}

    We create a second QuantumBool and evaluate some logical functions

    >>> q_bool_1 = QuantumBool()
    >>> print(q_bool_1 | q_bool_0)
    {False: 0.5, True: 0.5}
    >>> print(q_bool_1 & q_bool_0)
    {False: 1.0}

    QuantumBools are the results of comparisons:

    >>> from qrisp import QuantumFloat, QuantumChar
    >>> q_ch = QuantumChar()
    >>> q_ch[:] = {"g" : 1, "l" : -1}
    >>> q_bool_2 = (q_ch == "g")
    >>> q_bool_2.qs.statevector()
    sqrt(2)*(|g>*|True> - |l>*|False>)/2

    For :ref:`QuantumFloats <QuantumFloat>`, numeric comparison is also possible:


    >>> qf = QuantumFloat(4)
    >>> h(qf[3])
    >>> print(qf)
    {0: 0.5, 8: 0.5}
    >>> q_bool_3 = (qf >=  4)
    >>> print(q_bool_3)
    {False: 0.5, True: 0.5}

    To use a QuantumBool as a :ref:`ControlEnvironment`, we simply put it in a ``with``
    statement:

    ::

        with q_bool_3:
            qf += 2

    >>> print(qf)
    {0: 0.5, 10: 0.5}

    QuantumBools that are created directly after a ``with`` statement are uncomputed
    automatically:

    ::

        with qf == 10:
            q_bool_3.flip()

    >>> print(qf.qs)
    QuantumCircuit:
    ---------------
                       ┌────────────┐     ┌───────────┐
            qf.0: ─────┤0           ├─────┤0          ├──o─────────o──
                       │            │     │           │  │         │
            qf.1: ─────┤1           ├─────┤1          ├──■─────────■──
                       │            │     │  __iadd__ │  │         │
            qf.2: ─────┤2           ├─────┤2          ├──o─────────o──
                  ┌───┐│  less_than │     │           │  │         │
            qf.3: ┤ H ├┤3           ├─────┤3          ├──■─────────■──
                  └───┘│            │┌───┐└─────┬─────┘  │  ┌───┐  │
        lt_qbl.0: ─────┤4           ├┤ X ├──────■────────┼──┤ X ├──┼──
                       │            │└───┘               │  └─┬─┘  │
    lt_ancilla.0: ─────┤5           ├────────────────────┼────┼────┼──
                       └────────────┘                  ┌─┴─┐  │  ┌─┴─┐
      cond_env.0: ─────────────────────────────────────┤ X ├──■──┤ X ├
                                                       └───┘     └───┘
    Live QuantumVariables:
    ----------------------
    QuantumFloat qf
    QuantumBool lt_qbl

    Note that there is only a single QuantumBool listed in the "Live QuantumVariables"
    section, because the QuantumBool of the comparison
    ``qf == 10`` (called ``cond_env``) has been uncomputed.

    """

    def __init__(self, qs=None, name=None):
        QuantumVariable.__init__(self, 1, qs=qs, name=name)

        self.qfloat_comparison = False

    def decoder(self, integer):
        return bool(integer)

    def __and__(self, other):
        from qrisp import mcx

        and_qbl = QuantumBool()

        mcx(self.reg + other.reg, and_qbl.reg)

        return and_qbl

    def __or__(self, other):
        from qrisp import mcx, x

        or_qbl = QuantumBool()

        x(self)
        x(other)

        mcx(self.reg + other.reg, or_qbl.reg)

        x(self)
        x(other)

        x(or_qbl)

        return or_qbl

    def __xor__(self, other):
        from qrisp import cx

        xor_qbl = QuantumBool()

        cx(self, xor_qbl)
        cx(other, xor_qbl)

        return xor_qbl

    def flip(self):
        """
        Flips the QuantumBool's value.

        """

        from qrisp import x

        x(self)
        return self
    
    def __invert__(self):
        inverted_qbl = QuantumBool()
        from qrisp import cx
        cx(self, inverted_qbl)
        inverted_qbl.flip()
        return inverted_qbl
        

    def __enter__(self):
        self.env = control(self)
        self.env.__enter__()

    def __exit__(self, a, b, c):
        ref_count = sys.getrefcount(self)

        # If the refcount is 5, this means that the QuantumBool has been
        # created in a "with" statement and can therefore no longer be reached
        # after exiting. We therefore uncompute

        self.env.__exit__(a, b, c)
        if ref_count == 5:
            self.uncompute()
            
    def __bool__(self):
        raise Exception("Tried to convert QuantumBool to classical bool (probable due using the and + or keywords - try using & + | instead)")



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# Created by ann81984 at 29.04.2022
import numpy as np
import random
from qrisp import QuantumSession, QuantumVariable, p, cp
from qrisp.misc.GMS_tools import GXX_converter


def test_GMS_converter_example():
    n = 8

    qv = QuantumVariable(n)

    # Generate random phase-only circuit
    for i in range(n):
        for j in range(n):
            if random.randint(0, 1):
                if i == j:
                    p(np.pi / 2 ** random.randint(0, 3), qv[i])
                else:
                    cp(np.pi / 2 ** random.randint(0, 3), qv[i], qv[j])

    print(qv.qs)

    # Apply converter
    qc = GXX_converter(qv.qs)

    print("Is equal:", qv.qs.compare_unitary(qc, precision=4))

    unitary = qc.get_unitary()

    print(
        "Is diagonal:", np.linalg.norm(unitary - np.diag(np.diagonal(unitary))) < 1e-3
    )
    # print(np.round(np.angle(np.diagonal(qv.qs.get_unitary()))%(2*np.pi)/(2*np.pi), 3))
    assert qv.qs.compare_unitary(qc, precision=4)
    assert np.linalg.norm(unitary - np.diag(np.diagonal(unitary))) < 1e-3



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


# Modification to Kahns algorithm to reduce the depth of general circuits by applying
# trivial and non-trivial commutation relations based on the dag representation of
# unqomp.

import networkx as nx
import numpy as np
from numba import njit

from qrisp.uncomputation.unqomp import dag_from_qc


def qb_set_to_int(qubits, qc):
    res = 0
    for qb in qubits:
        res |= 1 << qc.qubits.index(qb)
    return res


# Kahns Algorithm based on
# https://www.geeksforgeeks.org/topological-sorting-indegree-based-solution/


def depth_sensitive_topological_sort(indices, indptr, int_qc, num_qubits):
    # Create a vector to store indegrees of all
    # vertices. Initialize all indegrees as 0.
    n = len(indptr) - 1
    in_degree = np.zeros(n, dtype=np.int32)

    depths = np.zeros(num_qubits, dtype=np.int32)
    # Traverse adjacency lists to fill indegrees of
    # vertices.  This step takes O(V + E) time

    for i in range(n):
        for j in indices[indptr[i] : indptr[i + 1]]:
            in_degree[j] += 1

    # Create a queue and enqueue all vertices with
    # indegree 0
    queue = []

    for i in range(n):
        if in_degree[i] == 0:
            queue.append(i)

    # Initialize count of visited vertices
    cnt = 0

    # Create a vector to store result (A topological
    # ordering of the vertices)
    top_order = np.zeros(n, dtype=np.int32)

    # One by one dequeue vertices from queue and enqueue
    # adjacents if indegree of adjacent becomes 0

    while queue:
        # The depth sensitive part is now to deque the node with the least depth
        node_costs = np.zeros(len(queue))

        for i in range(len(queue)):
            node = queue[i]

            qubits = int_qc[node]

            depth_sum = 0
            for j in range(num_qubits):
                if qubits & 1 << j:
                    depth_sum += depths[j]

            node_costs[i] = depth_sum

        u = queue.pop(np.argmin(node_costs))

        top_order[cnt] = u

        # Update depths array
        max_depth = 0

        for i in range(num_qubits):
            if int_qc[u] & 1 << i:
                if depths[i] > max_depth:
                    max_depth = depths[i]

        for i in range(num_qubits):
            if int_qc[u] & 1 << i:
                depths[i] = max_depth + 1

        # Update in degree array
        for i in indices[indptr[u] : indptr[u + 1]]:
            in_degree[i] -= 1
            if in_degree[i] == 0:
                queue.append(i)

        cnt += 1

    return top_order


depth_sensitive_topological_sort_jitted = njit(cache=True)(
    depth_sensitive_topological_sort
)


def reduce_depth(qc):
    if len(qc.data) <= 1:
        return qc.copy()

    dag = dag_from_qc(qc, remove_init_nodes=True)

    sprs_mat = nx.to_scipy_sparse_array(dag, format="csr")

    node_list = list(dag.nodes())
    qubit_ints = [qb_set_to_int(n.instr.qubits, qc) for n in node_list]

    try:
        qubit_ints = np.array(qubit_ints, dtype=np.int64)
        res = depth_sensitive_topological_sort_jitted(
            sprs_mat.indices, sprs_mat.indptr, qubit_ints, num_qubits=qc.num_qubits()
        )
    except OverflowError:
        res = depth_sensitive_topological_sort(
            sprs_mat.indices, sprs_mat.indptr, qubit_ints, num_qubits=qc.num_qubits()
        )

    qc_new = qc.clearcopy()

    for i in range(len(res)):
        qc_new.append(node_list[res[i]].instr)

    return qc_new



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import numpy as np

from qrisp.circuit.instruction import Instruction
from qrisp.circuit.operation import (
    Operation,
)
from qrisp.circuit.quantum_circuit import fast_append


# This function dissolves any Operation objects that have a definition circuit such
# that the result only consists of elementary gates
def transpile(qc, transpilation_level=np.inf, transpile_predicate=None, **kwargs):
    from qrisp.circuit import QuantumCircuit, Clbit, Qubit

    with fast_append():

        transpiled_qc = QuantumCircuit()

        # [transpiled_qc.add_qubit(Qubit(qb.identifier)) for qb in qc.qubits]
        # [transpiled_qc.add_clbit(Clbit(cb.identifier)) for cb in qc.clbits]

        for qb in qc.qubits:
            if isinstance(qb, Qubit):
                transpiled_qc.add_qubit(qb)
            else:
                transpiled_qc.add_qubit(Qubit(qb.identifier))

        for cb in qc.clbits:
            if isinstance(qb, Qubit):
                transpiled_qc.add_clbit(cb)
            else:
                transpiled_qc.add_clbit(Clbit(cb.identifier))

        translation_dic = {
            qc.qubits[i].identifier: transpiled_qc.qubits[i]
            for i in range(len(qc.qubits))
        }
        translation_dic.update(
            {
                qc.clbits[i].identifier: transpiled_qc.clbits[i]
                for i in range(len(qc.clbits))
            }
        )

        if transpile_predicate is None:
            transpile_predicate_ = lambda i, op: i < transpilation_level
        else:
            transpile_predicate_ = (
                lambda i, op: i < transpilation_level and transpile_predicate(op)
            )

        transpile_inner(qc, transpiled_qc, translation_dic, transpile_predicate_)

        QuantumCircuit.fast_append = False

        if not kwargs or not hasattr(qc, "to_qiskit"):
            return transpiled_qc
        else:
            from qrisp import QuantumCircuit

            qiskit_qc = transpiled_qc.to_qiskit()

            from qiskit import transpile as qiskit_transpile

            transpiled_qiskit_qc = qiskit_transpile(qiskit_qc, **kwargs)

            qrisp_qc = QuantumCircuit.from_qiskit(transpiled_qiskit_qc)

            return qrisp_qc


def transpile_inner(
        transpilation_qc,
        target_qc,
        translation_dic,
        transpile_predicate,
        transpilation_level=0,
):
    for i in range(len(transpilation_qc.data)):
        instr = transpilation_qc.data[i]
        if instr.op.definition:
            if transpile_predicate(transpilation_level, instr.op):
                definition = instr.op.definition

                new_translation_dic = {
                    definition.qubits[j].identifier: translation_dic[
                        instr.qubits[j].identifier
                    ]
                    for j in range(len(instr.qubits))
                }

                new_translation_dic.update(
                    {
                        definition.clbits[j].identifier: translation_dic[
                            instr.clbits[j].identifier
                        ]
                        for j in range(len(instr.clbits))
                    }
                )

                transpile_inner(
                    instr.op.definition,
                    target_qc,
                    new_translation_dic,
                    transpile_predicate,
                    transpilation_level + 1,
                )
                continue

        if not isinstance(instr.op, Operation):
            op = Operation(init_op=instr.op)
        else:
            op = instr.op

        target_qc.data.append(
            Instruction(
                op,
                [translation_dic[qb.identifier] for qb in instr.qubits],
                [translation_dic[cb.identifier] for cb in instr.clbits],
            )
        )
        # target_qc.append()


def extend(qc_0, qc_1, translation_dic="id"):
    if translation_dic == "id":
        translation_dic = {}
        for qb in qc_1.qubits:
            translation_dic[qb] = qb
            if qb not in qc_0.qubits:
                qc_0.add_qubit(qb)

        for cb in qc_1.clbits:
            translation_dic[qb] = qb
            if cb not in qc_0.clbits:
                qc_0.add_clbit(cb)

    # Copy in order to prevent modification
    translation_dic = dict(translation_dic)

    for key in list(translation_dic.keys()):
        if not isinstance(key, str):
            translation_dic[key.identifier] = translation_dic[key]

    for i in range(len(qc_1.data)):
        instruction_other = qc_1.data[i]
        qubits = []
        for qb in instruction_other.qubits:
            qubits.append(translation_dic[qb.identifier])

        clbits = []

        for cb in instruction_other.clbits:
            clbits.append(translation_dic[cb.identifier])

        instr_type = type(instruction_other)

        qc_0.data.append(instr_type(instruction_other.op, qubits, clbits))



# -*- coding: utf-8 -*-
"""
Created on Mon Jul  3 17:37:05 2023

@author: sea
"""

from qrisp.environments import QuantumEnvironment, GateWrapEnvironment
from qrisp.core.quantum_variable import QuantumVariable
from qrisp.core.compilation import qompiler
from qrisp.misc.utility import retarget_instructions
from qrisp.circuit import QubitAlloc

class IterationEnvironment(QuantumEnvironment):
    """
    This QuantumEnvironment can be used for reducing bottlenecks in compilation time. 
    Many algorithms such as Grover or QPE require repeated execution of the same
    quantum circuit. When scaling up complex algorithms that perform a lot of 
    non-trivial logic many iterations can significantly slow down the compilation
    speed. The ``IterationEnvironment`` remedies this flaw by recording the circuit
    of a single iteration and then duplicating the instructions the required 
    amount of times.
    
    Another bottleneck that can appear is the :meth:`compile <qrisp.QuantumSession.compile>` 
    method as the qubit allocation algorithm can also scale bad for really large
    algorithms. For this problem the ``IterationEnvironment`` exposes the ``precompile``
    keyword. Setting this keyword to ``True`` will perform the Qubit allocation algorithm
    on the QuantumEnvironments content and then (if necessary) allocate another
    :ref:`QuantumVariable` to accomodate the workspace qubits of the compilation 
    result. This way there is only a single (de)allocation per 
    ``IterationEnvironment``.
    
    .. note::
        
        Code that is executed within a ``IterationEnvironment`` may not 
        :meth:`delete <qrisp.QuantumVariable.delete>`
        previously created ``QuantumVariable`` and every created ``QuantumVariable``
        inside this :ref:`QuantumEnvironment` has to be deleted before exit.
    
    Parameters
    ----------
    
    qs : QuantumSession
        The ``QuantumSession`` in which the iterated code should be performed.
        QuantumVariables that have been created outside this ``QuantumEnvironment``
        need to be :ref:`registered <SessionMerging>` in this ``QuantumSession``.
    iteration_amount : integer
        The amount of iterations to perform.
    precompile : bool
        If set to ``True``, the qubit allocation algorithm will be run, once this
        ``QuantumEnvironment`` is compiled. This can significantly reduce the
        workload for a later compile call as there is only a single allocation
        per ``IterationEnvironment`` to handle.
        
    Examples
    --------
    
    We perform a simple addition circuit multiple times:
        
    ::
        
        from qrisp import QuantumFloat, IterationEnvironment
        
        a = QuantumFloat(5)
        
        with IterationEnvironment(a.qs, iteration_amount = 10):
            
            a += 1

    Evaluate the result
    
    >>> print(a)
    {10: 1.0}
    
    **Precompilation**
    
    Squaring a :ref:`QuantumFloat` uses the :meth:`qrisp.sbp_mult` function,
    which has a high demand of ancilla qubits.
    Therefore many iterations can quickly overload the allocation algorithm.
    
    ::
        
        def benchmark_function(use_iter_env = False):
            qf = QuantumFloat(5)
        
            if use_iter_env:
        
                with IterationEnvironment(qf.qs, 20, precompile = True):
        
                    temp = qf*qf
                    temp.uncompute()
            else:
        
                for i in range(20):
        
                    temp = qf*qf
                    temp.uncompute()
        
            compiled_qc = qf.qs.compile()
    
    
    Benchmark results:
        
    ::
        
        import time
        
        start_time = time.time()
        benchmark_function(False)
        print("Time taken without IterationEnvironment: ", time.time() - start_time)
        #Takes 55s
        
        start_time = time.time()
        benchmark_function(True)
        print("Time taken with IterationEnvironment: ", time.time() - start_time)
        #Takes 6s
    
    """
    
    
    def __init__(self, qs, iteration_amount, precompile = True):
        
        if iteration_amount < 1:
            raise Exception("Tried to create IterationEnvironment with < 1 iterations")
        
        self.iteration_amount = iteration_amount
        self.precompile = precompile
        
        QuantumEnvironment.__init__(self)
        
        self.env_qs = qs
        
        #Manual allocation management = False means that the compile function
        #pulls out all allocation gates to the front and all deallocation gates
        #to the back. This way the user doesn't have to worry about the (sometimes)
        #complicated logic of arranging them.
        #In this case we enable manual allocation management because the compilation
        #function is simple enough that we can ignore the allocation logic.
        self.manual_allocation_management = True
        
    
    
    def __enter__(self):
        
        self.inital_qvs = set(self.env_qs.qv_list)
        
        QuantumEnvironment.__enter__(self)
        
    def __exit__(self, exception_type, exception_value, traceback):
        
        if set(self.env_qs.qv_list) != self.inital_qvs:
            
            raise Exception("Tried to invoke IterationEnvironment with code creating/deleting QuantumVariables")
        
        QuantumEnvironment.__exit__(self, exception_type, exception_value, traceback)
        
        
    def compile(self):
        
        #Stow away the environment data to facicility environment compilation
        temp_qs_data = list(self.env_qs.data)
        self.env_qs.data = []
        
        #If the code executed in the Iteration environment contains many 
        #(de)allocations the allocation algorithm in the compile method can 
        #be a bottleneck.
        #The precompilation feature calls the compile method on the quantum session
        #and appends the compiled data instead. This way there is only a single
        #(de)allocation for an arbitrary amount of iterations.
        #This comes at the cost that the allocation algorithm might find better
        #ways if it has insight into the internal allocation structure.
        if self.precompile:
            
            # # #We append the previously executed allocation calls such that
            # # #the compile method knows which variables are allocated
            for qv in self.env_qs.qv_list:
                for qb in qv.reg:
                    self.env_qs.append(QubitAlloc(), qb)
            
            self.env_qs.barrier(self.env_qs.qubits)
            
            #Compile the quantum environment to retrieve the compiled data
            QuantumEnvironment.compile(self)
            
            compiled_qc = self.env_qs.compile()
            
            #Remove the (de)allocation calls from the compiled quantum circuit
            compiled_data = []
            for instr in compiled_qc.data:
                if "alloc" not in instr.op.name or instr.op.name == "barrier":
                    compiled_data.append(instr)
            
            #Reinstate the original data
            self.env_qs.data = temp_qs_data
            
            
            #The actual compilation begins now
            
            #Determine the workspace qubits from the compiled qc
            workspace_qubits = list(set(compiled_qc.qubits) - set(self.env_qs.qubits))
            
            #Allocate a QuantumVariable that will hold the workspace
            workspace_var = QuantumVariable(len(workspace_qubits), qs = self.env_qs, name = "workspace_var*")
            
            
            #We now prepare the qubit lists for the retarget_instructions function
            
            env_qubit_names = [qb.identifier for qb in self.env_qs.qubits]
            
            source_qubits = []
            target_qubits = []

            for i in range(len(compiled_qc.qubits)):
                try:
                    qubit_index = env_qubit_names.index(compiled_qc.qubits[i].identifier)
                    target_qubits.append(self.env_qs.qubits[qubit_index])
                    source_qubits.append(compiled_qc.qubits[i])
                except ValueError:
                    source_qubits.append(workspace_qubits.pop())
                    target_qubits.append(workspace_var[len(workspace_qubits)])
            
            #Perform instruction retargeting
            retarget_instructions(compiled_data, source_qubits, target_qubits)

            #Perform iterated instruction execution            
            for i in range(self.iteration_amount):
                compiled_data = [instr.copy() for instr in compiled_data]                
                self.env_qs.data.extend(compiled_data)
                
            #Delete workspace variable
            workspace_var.delete()

        #The non-precompiled case is much simpler
        else:
            
            QuantumEnvironment.compile(self)
            
            compiled_data = list(self.env_qs.data)
            
            self.env_qs.data = temp_qs_data
            
            for i in range(self.iteration_amount):
                compiled_data = [instr.copy() for instr in compiled_data]
                self.env_qs.data.extend(compiled_data)
                

            
        
        
        
        


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# -*- coding: utf-8 -*-

import threading

import numpy as np

from qrisp.simulator.quantum_state import QuantumState

thread_limit = 1


# This class describes a set of decoherent quantum states
class ImpureQuantumState:
    # Initially we start with a single quantum state
    def __init__(self, n, clbit_amount=0):
        self.states = [QuantumState(n)]
        self.cl_prob = [1]
        self.outcome_list = [np.zeros(clbit_amount, dtype=np.byte)]
        self.collected_measurements = []

    def apply_operation(self, operation, qubits):
        # Perform operation application in parallel. This is possible because many
        # functions in here release the GIL (such as numpy's np.dot)
        if thread_limit > 1 and len(self.states) > 1:
            threads = []

            for i in range(len(self.states)):
                mp_wrapper = lambda: self.states[i].apply_operation(operation, qubits)

                thr = threading.Thread(target=mp_wrapper)

                if len(threads) >= thread_limit:
                    threads[0].join()
                    threads.pop(0)

                thr.start()
                threads.append(thr)

            [thr.join() for thr in threads]

        else:
            for i in range(len(self.states)):
                self.states[i].apply_operation(operation, qubits)

    def conditionally_apply_operation(self, operation, qubits, clbit):
        # Perform operation application in parallel. This is possible because many
        # functions in here release the GIL (such as numpy's np.dot)

        if thread_limit > 1:
            threads = []

            for i in range(len(self.states)):
                if self.outcome_list[clbit] == "0":
                    continue

                mp_wrapper = lambda: self.states[i].apply_operation(operation, qubits)

                if len(threads) >= thread_limit:
                    threads[0].join()
                    threads.pop(0)

                thr = threading.Thread(target=mp_wrapper)
                thr.start()

                threads.append(thr)

            [thr.join() for thr in threads]

        else:
            for i in range(len(self.states)):
                if self.outcome_list[clbit] == "0":
                    continue

                self.states[i].apply_operation(operation, qubits)

    # Perform reset operation
    # We perform the reset operation by measuring but not tracking the measurement
    # result.
    def reset(self, qubit_index, keep_outcome=False):
        if thread_limit > 1:
            threads = []

            for i in range(len(self.states)):
                mp_wrapper = lambda: self.states[i].measure(qubit_index, keep_outcome)
                thr = threading.Thread(target=mp_wrapper)

                if len(threads) > thread_limit:
                    threads[0].join()
                    threads.pop(0)

                thr.start()
                threads.append(thr)

        else:
            for i in range(len(self.states)):
                self.states[i].measure(qubit_index, keep_outcome)

        # Generate new list of states and the corresponding classical propabilities +
        # the outcome bitstrings
        new_states = []
        new_cl_prob = []
        new_outcome_list = []

        for i in range(len(self.states)):
            # Wait for measurement threads to finish
            if thread_limit > 1:
                if len(threads):
                    threads[0].join()
                    threads.pop(0)

            # Get measurement outcome
            # p_0 represents the probability to measure a 0,
            # state_0 represents the state after measuring a 0
            p_0, state_0, p_1, state_1 = self.states[i].last_mes_outcome
            # Log results
            if p_0 != 0:
                # Add the new decoherent state to the list of states
                new_states.append(state_0)

                # Update its classical probability
                new_cl_prob.append(p_0 * self.cl_prob[i])

                # The bitstring stays the same
                new_outcome = self.outcome_list[i].copy()
                new_outcome_list.append(new_outcome)

            # Perform the same thing for the outcome 1
            if p_1 != 0:
                new_states.append(state_1)
                new_cl_prob.append(p_1 * self.cl_prob[i])
                new_outcome = self.outcome_list[i].copy()
                new_outcome_list.append(new_outcome)

        # Update attributes
        self.states = new_states
        self.cl_prob = new_cl_prob
        self.outcome_list = new_outcome_list

    # This functions performs measurements
    # The logic behind it is rather similar to reset, but this time we log the
    # measurement outcome.
    def measure(self, qubit_index, clbit_index):
        if thread_limit > 1:
            threads = []

            for i in range(len(self.states)):
                mp_wrapper = lambda: self.states[i].measure(qubit_index)
                thr = threading.Thread(target=mp_wrapper)

                if len(threads) > thread_limit:
                    threads[0].join()
                    threads.pop(0)

                thr.start()
                threads.append(thr)

        else:
            for i in range(len(self.states)):
                self.states[i].measure(qubit_index, True)

        new_states = []
        new_cl_prob = []
        new_outcome_list = []

        for i in range(len(self.states)):
            # Wait for measurement threads to finish
            if thread_limit > 1:
                if len(threads):
                    threads[0].join()
                    threads.pop(0)

            p_0, state_0, p_1, state_1 = self.states[i].last_mes_outcome

            if p_0 != 0:
                new_states.append(state_0)
                new_cl_prob.append(p_0 * self.cl_prob[i])
                new_outcome = self.outcome_list[i].copy()

                # Here is the point, where we log the measurement outcome
                new_outcome[clbit_index] = 0
                new_outcome_list.append(new_outcome)
            if p_1 != 0:
                new_states.append(state_1)
                new_cl_prob.append(p_1 * self.cl_prob[i])
                new_outcome = self.outcome_list[i].copy()
                new_outcome[clbit_index] = 1
                new_outcome_list.append(new_outcome)

        self.states = new_states
        self.cl_prob = new_cl_prob
        self.outcome_list = new_outcome_list

    def add_qubit(self):
        for state in self.states:
            state.add_qubit()

    def add_clbit(self):
        for i in range(len(self.outcome_list)):
            self.outcome_list[i] = np.array(
                [0] + list(self.outcome_list[i]), dtype=np.byte
            )



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import itertools

import numpy as np

from qrisp.core import QuantumVariable, recursive_qv_search
from qrisp.logic_synthesis import TruthTable
from qrisp.misc import bin_rep, custom_qv, int_as_array


class QuantumDictionary(dict):
    r"""
    This class can be used for loading data relations into the quantum computer which
    are not based on a quantum algorithm.

    As an inheritor of the Python dictionary it has all the functionality we are used to

    >>> from qrisp import QuantumDictionary, QuantumVariable, multi_measurement
    >>> qd = QuantumDictionary()
    >>> qd[1] = 2
    >>> qd[42] = (3,4)
    >>> qd["hello"] = "hallo"
    >>> qd["world"] = "welt"
    >>> print(qd[42])
    (3,4)

    The key difference is that the QuantumDictionary can also recieve
    :ref:`QuantumVariables <QuantumVariable>` as keys and return the corresponding
    values as an entangled QuantumVariable.

    We demonstrate this by preparing a QuantumVariable which has the keys of ``qd`` as
    outcome labels:

    >>> key_qv = QuantumVariable.custom([1, 42, "hello", "world"])
    >>> key_qv[:] = {"hello" : 1, "world" : 1}
    >>> print(key_qv)
    {'hello': 0.5, 'world': 0.5}

    We now load the data from the QuantumDictionary into the quantum algorithm by
    dereferencing ``qd`` with ``key_qv``:

    >>> value_qv = qd[key_qv]
    >>> multi_measurement([key_qv, value_qv])
    {('hello', 'hallo'): 0.5, ('world', 'welt'): 0.5}

    We see that the states of ``key_qv`` are now entangled with the states of the
    values.

    QuantumDictionaries can also load from tuples of QuantumVariables:

    >>> qd[(1,2)] = 73
    >>> qd[(0,2)] = 37
    >>> qf1 = QuantumFloat(1)
    >>> qf2 = QuantumFloat(2)
    >>> h(qf1)
    >>> qf2[:] = 2
    >>> res = qd[(qf1, qf2)]
    >>> multi_measurement([qf1, qf2, res])
    {(0, 2, 37): 0.5, (1, 2, 73): 0.5}

    Mathematically we have

    .. math::

        U_{\text{qd}} \left( \sum_{x \in \text{labels}} a_x \ket{x} \right)
        \ket{0} = \sum_{x \in \text{labels}} a_x \ket{x} \ket{\text{qd}[x]}

    Note that this quantum operation is realized through quantum logic synthesis, which
    scales rather badly compared to algorithmic generation of data relations.
    Therefore, performing as much logic on the quantum computer is preferable over
    performing the logic on the classical computer and inserting the results using
    QuantumDictionaries.


    **Specifying the return type**

    The returned QuantumVariable ``value_qv`` is (similar to ``key_qv``) a
    CustomQuantumVariable:

    >>> print(type(value_qv))
    <class 'qrisp.misc.misc_functions.custom_qv.<locals>.CustomQuantumVariable'>

    If we want to apply further processing this might not be helpfull since custom
    QuantumVariables lack many methods that are available in more specific quantum
    types. In this case we can supply the QuantumDictionary with a return type:

    >>> from qrisp import QuantumFloat
    >>> qtype = QuantumFloat(4, -2, signed = True)
    >>> float_qd = QuantumDictionary(return_type = qtype)

    We fill again with some example values

    >>> float_qd["hello"] = 0.5
    >>> float_qd["world"] = -1

    And retrieve the value:

    >>> value_qv = float_qd[key_qv]
    >>> print(type(value_qv))
    <class 'qrisp.qtypes.quantum_float.QuantumFloat'>

    Since ``value_qv`` is a QuantumFloat now, we can use the established methods for
    arithmetic - for instance the inplace addition:

    >>> value_qv += 1.5
    >>> print(multi_measurement([key_qv, value_qv]))
    {('hello', 2.0): 0.5, ('world', 0.5): 0.5}


    **Advanced usage**

    In some cases, (such as manual uncomputation) it is required to specify into which
    variable the QuantumDictionary should load. We do this with the
    :meth:`load <qrisp.QuantumDictionary.load>` method:

    >>> qf = qtype.duplicate()
    >>> float_qd.load(key_qv, qf)
    >>> print(qf)
    {0.5: 0.5, -1.0: 0.5}

    The ``load`` method furthermore allows to specify which logic synthesis algorithm
    should be used.

    >>> qf2 = qtype.duplicate()
    >>> float_qd.load(key_qv, qf2, synth_method = "gray")

    """

    def __init__(self, init_dict={}, return_type=None):
        super().__init__(init_dict)

        self.return_type = return_type

    def __getitem__(self, key):
        return self.load(key)

    def load(self, key, value_qv=None, synth_method="gray"):
        """
        Loads the values of the QuantumDictionary into a given QuantumVariable.

        Parameters
        ----------
        key_qv : QuantumVariable
            A QuantumVariable with a decoder supporting the keys of this
            QuantumDictionary.
        value_qv : QuantumVariable, optional
            The QuantumVariable to load the values into. If given None, a new
            QuantumVariable is generated.
        synth_method : string, optional
            The method of logic synthesis to use for loading. Currently available are
            "gray", "gray_pt", "pprm" and "pprm_pt". The default is "gray".

        Raises
        ------
        Exception
            Tried to load value from empty dictionary.

        Examples
        --------

        We create a QuantumDictionary with return type QuantumFloat

        >>> from qrisp import QuantumDictionary, QuantumFloat, h
        >>> qtype = QuantumFloat(4, -2)
        >>> float_qd = QuantumDictionary(return_type = qtype)
        >>> float_qd[0] = 1
        >>> float_qd[1] = 2

        Create the key and the value variable:

        >>> key_qv = QuantumFloat(1, signed = False)
        >>> h(key_qv)
        >>> value_qv = qtype.duplicate()

        And load the values

        >>> float_qd.load(key_qv, value_qv, synth_method = "pprm")
        >>> print(value_qv)
        {1.0: 0.5, 2.0: 0.5}

        """

        qv_list = recursive_qv_search(key)
        if not len(qv_list):
            return dict.__getitem__(self, key)

        if not isinstance(key, tuple):
            key = (key,)

        labels = []

        for qv in key:
            if not isinstance(qv, QuantumVariable):
                raise Exception(
                    "Tried to deref QuantumDictionary with mixed"
                    "(classical + quantum) input."
                )

            labels.append([qv.decoder(i) for i in range(2**qv.size)])

        quantum_key = key

        if len(labels) == 1:
            constellations = labels[0]
        else:
            constellations = itertools.product(*labels)

        relevant_dic = {}

        for const in constellations:
            if const in self:
                relevant_dic[const] = dict.__getitem__(self, const)

        if value_qv is None:
            if isinstance(self.return_type, type(None)):
                value_qv = custom_qv(list(relevant_dic.values()))
            else:
                value_qv = self.return_type.duplicate()

        n = sum([qv.size for qv in key])

        tt_array = np.zeros((2**n, value_qv.size))

        for k in relevant_dic:
            bin_string = ""

            for i in range(len(quantum_key)):
                qv = quantum_key[i]

                if isinstance(k, tuple):
                    bin_string += bin_rep(qv.encoder(k[i]), qv.size)[::-1]
                else:
                    bin_string += bin_rep(qv.encoder(k), qv.size)[::-1]

            row_number = int(bin_string[::-1], 2)
            tt_array[row_number, :] = int_as_array(
                value_qv.encoder(relevant_dic[k]), value_qv.size
            )[::-1]

        tt = TruthTable(tt_array)

        qv_temp_0 = QuantumVariable(n)
        qv_temp_1 = QuantumVariable(value_qv.size, qs=qv_temp_0.qs)

        tt.q_synth(qv_temp_0, qv_temp_1, method=synth_method)

        if len(qv_temp_0.qs.qubits) != qv_temp_0.size + qv_temp_1.size:
            synth_ancilla = QuantumVariable(
                len(qv_temp_0.qs.qubits) - qv_temp_0.size + qv_temp_1.size
            )
            quantum_key.append(synth_ancilla)

        res_gate = qv_temp_0.qs.to_gate("q_load")
        res_gate.is_qfree = True
        res_gate.permeability = {i: i < n for i in range(res_gate.num_qubits)}

        quantum_key[0].qs.append(
            res_gate, sum([qv.reg for qv in quantum_key], []) + value_qv.reg
        )

        if len(qv_temp_0.qs.qubits) != qv_temp_0.size + qv_temp_1.size:
            synth_ancilla.delete()

        return value_qv



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# -*- coding: utf-8 -*-

from qrisp.simulator.bi_arrays import DenseBiArray, SparseBiArray, tensordot
from qrisp.simulator.numerics_config import float_tresh, xp

# tensordot = xp.tensordot

# Class to describe TensorFactors


class TensorFactor:
    def __init__(self, qubits, init_tensor_array=None):
        # This list contains a list of integers, which describe the current permutation
        # of qubit indices inside this TensorFactor, i.e. swapping two elements of this
        # list is the same as applying a swap gate. Using this structure is advantageous
        # because applying matrices on specific qubit indices requires alot of swapping.
        # Swapping the indices of an array is ressource costly because the array needs
        # to be copied. As it turns out memory speed is a bottleneck.
        self.qubits = qubits

        self.n = len(qubits)

        if init_tensor_array is None:
            self.tensor_array = xp.zeros(2**self.n, dtype=xp.complex64)
            self.tensor_array[0] = 1

        else:
            self.tensor_array = init_tensor_array

        if isinstance(self.tensor_array, xp.ndarray):
            self.tensor_array = SparseBiArray(self.tensor_array)

    def copy(self):
        return TensorFactor(list(self.qubits), self.tensor_array.copy())

    # Swaps two qubit indices
    def swap(self, i, j):
        if i == j:
            return

        self.qubits[i], self.qubits[j] = self.qubits[j], self.qubits[i]

        self.tensor_array = self.tensor_array.reshape(self.n * [2])
        self.tensor_array = self.tensor_array.swapaxes(i, j)
        self.tensor_array = self.tensor_array.reshape(2**self.n)

    # Applies a matrix onto the given state
    def apply_matrix(self, matrix, qubits):
        # Remove the non-zero entries caused by floating point errors to get
        # the most out of the sparse matrix representation
        import numpy as np

        if matrix.size >= 2**6:
            if matrix.dtype == np.complex64:
                matrix = matrix * (np.abs(matrix) > 1e-7)
            if matrix.dtype == np.complex128:
                matrix = matrix * (np.abs(matrix) > 1e-15)

        # Convert matrix to BiArray
        matrix = DenseBiArray(matrix)

        # Swap the qubits which the matrix is supposed to be applied on to the front
        for i in range(len(qubits)):
            qubit_index = self.qubits.index(qubits[i])
            self.swap(qubit_index, i)

        # Reshape tensor array such that the qubits form a single index
        reshaped_tensor_array = self.tensor_array.reshape(
            (matrix.shape[0], 2**self.n // matrix.shape[0])
        )

        # Apply matrix
        self.tensor_array = tensordot(
            matrix, reshaped_tensor_array, axes=(1, 0), contract_sparsity_threshold=0.05
        ).reshape(2**self.n)
        # print(type(self.tensor_array))

    # Entangles two TensorFactors.
    def entangle(self, other):
        if self == other:
            return self
        if set(self.qubits).intersection(other.qubits):
            raise Exception(
                "Tried to entangle two tensor factors with overlapping qubits"
            )

        # The tensordot function with axes = ((),()) performs a,b -> a (x) b, where
        # (x) is the tensor product sign
        init_array = tensordot(
            other.tensor_array,
            self.tensor_array,
            axes=((), ()),
            contract_sparsity_threshold=0.01,
        ).ravel()

        return TensorFactor(other.qubits + self.qubits, init_array)

    # Perform measurement
    def measure(self, i):
        # Swap the index that is supposed to be measured to the front
        index = self.qubits.index(i)
        self.swap(0, index)

        # Split the array - the lower half corresponds to the state with outcome 0,
        # the upper half to 1
        lower_half, upper_half = self.tensor_array.split()

        # Determine probabilities
        p_0 = lower_half.squared_norm()
        p_1 = upper_half.squared_norm()

        # p_1 = abs(p_0 - 1) #This doesn't work due to floating point errors
        # accumulating in the normalization

        # Prepare the qubit list of the resulting TensorFactor
        # After the measurement, the measured qubit is disentangled from the rest,
        # so we can put it into it's own TensorFactor object
        new_qubit_list = list(self.qubits)
        new_qubit_list.pop(0)

        # This treats the case that both measurement probabilities are non-zero
        if p_0 > float_tresh and p_1 > float_tresh:
            # Normalize the new statevector arrays
            normalization = 1 / (p_0**0.5)
            lower_half.data *= normalization

            # Create tensor factor corresponding to outcome 0
            tensor_factor_0 = TensorFactor(new_qubit_list, lower_half)

            # Perform the same operation for outcome 1
            normalization = 1 / (p_1**0.5)
            upper_half.data *= normalization

            tensor_factor_1 = TensorFactor(list(new_qubit_list), upper_half)

        # This treats the cases that one of the outcomes has 0 probability
        elif p_0 < float_tresh:
            p_0 = 0
            p_1 = 1
            tensor_factor_0 = None
            tensor_factor_1 = TensorFactor(new_qubit_list, upper_half)

        else:
            p_0 = 1
            p_1 = 0
            tensor_factor_0 = TensorFactor(new_qubit_list, lower_half)
            tensor_factor_1 = None

        return p_0, tensor_factor_0, p_1, tensor_factor_1

    def unravel(self):
        sorted_qubit_list = list(self.qubits)
        sorted_qubit_list.sort()
        for i in range(len(self.qubits)):
            if sorted_qubit_list[i] == self.qubits[i]:
                continue
            self.swap(i, self.qubits.index(sorted_qubit_list[i]))


# This function allow entangling more than two TensorFactors.
# Since we are free to choose a contraction order, we pick one
# that synergizes with the parallelization features of the BiArrays,
# i.e. we entangle two factors, and put the entangled result at
# the back of a queue. The processing of the entangling is hopefully
# done once this factor is at the front of the queue.
def multi_entangle(input_factor_list):
    # If there is only one factor, there is nothing to be done
    if len(input_factor_list) == 1:
        return input_factor_list[0]

    else:
        # In the case of multiple factors, entangle the first two, and put the result
        # at the back

        input_factor_list.sort(key=lambda x: len(x.qubits))
        input_factor_list.append(input_factor_list[0].entangle(input_factor_list[1]))

        # Remove entangling factors from the queue
        input_factor_list.pop(0)
        input_factor_list.pop(0)

        # Recurse on remaining factors
        return multi_entangle(input_factor_list)



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# -*- coding: utf-8 -*-

import threading
import time

import numpy as np
from scipy.sparse import (
    bsr_matrix,
    coo_array,
    coo_matrix,
    csc_array,
    csc_matrix,
    csr_array,
    csr_matrix,
)

import qrisp.simulator.bi_array_helper as hlp

try:
    # sparse_dot_mkl seems to be only faster in situations, where the shape of the
    # matrices, results in better parallelization. For many quantum circuits, this was
    # not the case so this feature is disabled until further notice.
    raise Exception
    # Install command: conda install -c conda-forge sparse_dot_mkl
    from sparse_dot_mkl import dot_product_mkl

    def sparse_matrix_mult(a, b):
        if not isinstance(a, (np.ndarray, csr_array, csc_array)):
            a = a.tocsr()
        if not isinstance(b, (np.ndarray, csc_array, csr_array)):
            b = b.tocsc()

        return dot_product_mkl(a, b)

except:
    # print("Failed to import mkl sparse matrix multiplication. Install:
    # with conda install -c conda-forge sparse_dot_mkl. Using scipy algorithm.")
    sparse_matrix_mult = lambda a, b: (a @ b)


# A quick helper function to evaluate the product of an iterable
def prod(iter):
    res = 1
    for i in range(len(iter)):
        res *= iter[i]
    return res


class BiArray:
    def v_dot(self, other):
        original_shape_self = self.shape
        orignal_shape_other = other.shape

        self.reshape((self.size,))
        other.reshape((other.size,))

        res = tensordot(self, other, (0, 0))

        self.reshape(original_shape_self)
        other.reshape(orignal_shape_other)

        return res


# The classes inside this file aim describe so called BiArrays.
# A BiArray is a nd-array, where all the dimensions is are a power of 2,
# i.e. 2, 4, 8, or 1024... This property can be used to efficiently manipulate indices
# of sparse matrices such that they can be processed quickly enough to use sparse matrix
# multiplication libraries for tensor contraction. This functionality is summarized in
# the SparseBiArray class. The DenseBiArray class is an interface to numpy arrays,
# that also use the restriction of the dimension being a power of two. For instance does
# the DenseBiArray class support constant time reshaping and axis swapping. The
# reshaping and axis swappings are "collected" and executed together once the array is
# required for contraction.

# Both classes automatically make use of potential parallelism in the code.
# This is achieved by returning arrays where the content is "under construction" by a
# certain thread. If the "under construction" array is required in a contraction, the
# finishing of the calculation process is awaited.

# The tensordot function is a numpy-esque interface for performing tensor contractions.
# This function also automatically converts between sparse and dense BiArrays once a
# specified sparsity threshold is reached.

# A third class, the DummyBiArray, is a class that provides the same methods as the
# previous two, however the contractions are not actually executed, but instead the
# amount of floating point operations is logged in a class attribute. This allows to
# quickly evaluate the performance of a contraction order without actually contracting.


multithreading_threshold = np.inf

# This SparseBiArray class provides a numpy-esque rank n tensor interface, which uses
# sparse matrix multiplication for tensor contraction. The fact that only powers of two
# are allowed for the size, allow efficient index manipulations, that permit faster
# construction and processing of said sparse matrices compared to for instance the
# rank n tensor interface of the "sparse" package.

# The idea is to store multi-dimensional sparse arrays as the coo-array of their raveled
# version i.e. we ravel the array and store the indices of the non-zero entries together
# with their data. So for instance the array
# array([[0, 2, 0, 0],
#        [3, 0, 0, 0]])
# is stored as with the non-zero indices nz_indices = [1, 4] and data = [2,3].
# If we now want to swap the axes of the array in this representation, we can do
# bit-level manipulations on the nz_indices array:
# In binary we have
# nz_indices = ["001", "100"]
# An axis swap can now be performed by manipulating the binary strings (which is
# possible rather efficiently). In this case the manipulation consists of moving the
# last two digits of the bitstring to the front. After the axis swap we have
# nz_indices_axis_swapped = ["010", "001"] = [2, 1]
# If we now unravel the array with updated shape, we get
# array([[0, 3],
#        [2, 0],
#        [0, 0],
#        [0, 0]])
# Which confirms that this bit-level operation indeed represented an axis-swap


class SparseBiArray(BiArray):
    # This attribute decides after how many contractions, there should be a check for
    # duplicates and zero entries in the sparse matrix data
    contraction_counter_threshold = np.inf

    # Constructor of the SparseBiArray class.
    # init_object can be
    # - a numpy array
    # - a tuple of arrays where init_object[0] is an array of non-zero indices
    #     and init_object[1] the corresponding data
    # - a sparse array in coo format
    def __init__(self, init_object, shape=None, contraction_counter=0):
        # This attribute tracks how many sparse matrix contractions, this object has
        # been through
        self.contraction_counter = contraction_counter

        # Handle the case of an array
        if isinstance(init_object, np.ndarray):
            array = init_object

            # Store shape
            self.shape = array.shape

            # Ravel init array
            raveled_array = array.ravel()

            # raveled_array = np.around(raveled_array, 6, out = raveled_array)

            # Find non-zero indices
            self.nz_indices = np.nonzero(raveled_array)[0]

            # Store data
            self.data = raveled_array[self.nz_indices]

            # self.nz_indices, self.data = hlp.elim_zeros(self.nz_indices, self.data)

        # Handle the case that the init_object is a tuple of nz_indices/data arrays
        elif isinstance(init_object, tuple):
            if shape is None:
                raise Exception(
                    "Tried to initialise SparseBiArray from sparse data without "
                    "providing a shape"
                )

            self.nz_indices = init_object[0].ravel()
            self.data = init_object[1]

            self.nz_indices, self.data = hlp.elim_zeros(self.nz_indices, self.data)

        # Handle the case that the init_object is a sparse matrix/array
        elif isinstance(init_object, (coo_array, coo_matrix)):
            if shape is None:
                raise Exception(
                    "Tried to initialise SparseBiArray from sparse matrix without "
                    "providing a shape"
                )

            self.nz_indices = hlp.gen_flat_coords(
                init_object.col, init_object.row, init_object.shape
            )
            self.data = init_object.data

        elif isinstance(init_object, SparseBiArray):
            copy = init_object.copy()

            for var in copy.__dict__.keys():
                self.__dict__[var] = copy.__dict__[var]

            return
        else:
            raise Exception(
                "Tried to initialize SparseBiArray with unknown type "
                + str(type(init_object))
            )
        if shape is not None:
            self.shape = shape

        # Store the size
        self.size = prod(self.shape)

        if not ((self.size & (self.size - 1) == 0) and self.size != 0):
            raise Exception(
                "Tried to initialize SparseBiArray with size "
                + str(self.size)
                + " which is not a power of 2"
            )

        # Store the sparsity
        self.sparsity = len(self.nz_indices) / self.size

        # This list keeps track of the axes permutations that have been applied.
        # Once the SparseBiArray is needed for a contraction, all these swaps are
        # applied within a single function call and then the corresponding matrix is
        # build from that.
        self.index_bit_permutation = list(range(int(np.log2(self.size))))

        self.dtype = self.data.dtype

        # If the contraction counter is above the threshold, we sum the duplicates and
        # eliminate the zeros.
        if False:
            # if contraction_counter > self.contraction_counter_threshold:
            self.sum_duplicates()
            self.eliminate_zeros()
            self.contraction_counter = 0

    # Method for copying
    def copy(self):
        self.apply_swaps()
        return SparseBiArray(
            (self.nz_indices.copy(), self.data.copy()), shape=tuple(self.shape)
        )

    # This method checks if the sparse array has a thread object, which means that
    # it's data is under construction. If so, it waits until the thread is finished
    def catch_up(self):
        if hasattr(self, "thread"):
            start_time = time.time()
            self.thread.join()

            duration = time.time() - start_time
            if duration > 0.001:
                # print(duration)
                pass
            del self.thread

    # Reshaping method - note that no data is manipulated, i.e. constant time
    def reshape(self, shape):
        if isinstance(shape, int):
            shape = (shape,)

        shape = tuple(shape)
        if prod(shape) != prod(self.shape):
            raise Exception("Tried to reshape with invalid shape")

        self.shape = shape

        return self

    # Method for swapping two axes
    def swapaxes(self, i, j):
        if i == j:
            return
        if i > j:
            i, j = j, i

        # In order to store the swap in the index_bit_permutation attribute,
        # we need to swap the corresponding digit ranges inside this list

        # For this we first determine the digit range of the index bitstrings
        # For instance imagine we have the index 215 and the shape (4,8,8)
        # and we want to swap the axes i = 0 and j = 2.
        # In binary, 215 is 11010111.
        # We first determine the layout of the indices inside this bitstring
        bit_shape = np.log2(self.shape).astype(np.int64)

        # In our case the bit shape is (2,3,3)

        i_interval_start = np.sum(bit_shape[:i])
        i_interval_end = i_interval_start + bit_shape[i]

        # In our example i = 0, the i_interval starts at the first (most significant)
        # digit and ends after the third, i.e. |11|010111

        j_interval_start = np.sum(bit_shape[:j])
        j_interval_end = j_interval_start + bit_shape[j]

        # The j interval now starts at 6-th digit and ends at the last digit
        # ie. |11|010|111|

        # these two intervals have to be swapped.
        # However, in order to make this method constant in time, the actual swapping
        # of the indices is not performed here - rather we log the permutation of the
        # indices.

        # After creating of our SparseBiArray the attribute .index_bit_permutation
        # simply holds the numbers from 0 to log2(size)-1, i.e. in the case of our
        # example [0,1,2,3,4,5,6,7]
        # In order to log the axis swap, we perform this permutation on the
        # .index_bit_permutation attribute, i.e. after the application of this method,
        # we want it to be [5,6,7,2,3,4,0,1]
        # i.e. the interval of the first two and the last three swapped

        # In order to do this we split the index_bit_permutation list into 5 parts
        # 1. the interval before the i-interval
        # 2. the i-interval
        # 3. the interval between i and j
        # 4. the j interval
        # 5. the interval after j

        # in the case of our example, we would have
        # 1. = []
        # 2. = [0,1]
        # 3. = [2,3,4]
        # 4. = [5,6,7]
        # 5. = []

        # We than acquire the permuted list by concatenating these list in the correct
        # order, i.e. 1. + 4. + 3. + 2. + 5.

        # Determine the first interval
        init_interval = self.index_bit_permutation[:i_interval_start]

        # Determine the i-interval
        i_interval = self.index_bit_permutation[i_interval_start:i_interval_end]

        # Determine the middle-interval
        middle_interval = self.index_bit_permutation[i_interval_end:j_interval_start]

        # Determine the j-interval
        j_interval = self.index_bit_permutation[j_interval_start:j_interval_end]

        # Determine the final interval
        final_interval = self.index_bit_permutation[j_interval_end:]

        # Concatenate intervals
        self.index_bit_permutation = (
            init_interval + j_interval + middle_interval + i_interval + final_interval
        )

        # Create the tuple for the new shape
        new_shape = list(self.shape)

        new_shape[i], new_shape[j] = new_shape[j], new_shape[i]

        self.shape = tuple(new_shape)

        return self

    # This method applies the index permutation generated too
    def apply_swaps(self):
        # Wait for any contraction processes to finish
        self.catch_up()

        # Check if any permutations have to be performed
        if self.index_bit_permutation == list(range(int(np.log2(self.size)))):
            return

        # Apply permutations
        # The bit level manipulations are done in the helper file
        self.nz_indices = hlp.permute_axes(
            self.nz_indices, np.array(self.index_bit_permutation)
        ).ravel()

        # Reset the index_bit_permutation attribute, to
        # indicate that no permutation has to be executed now
        self.index_bit_permutation = list(range(int(np.log2(self.size))))

    # This method sums duplicates ie. if an index appears twice in nz_indices,
    # the corresponding data entries are summed
    def sum_duplicates(self):
        self.nz_indices, self.data = hlp.sum_duplicates(self.nz_indices, self.data)

    # If after a contraction a new zero appeared, this entry does not need to tracked
    # anymore. This method removes the zeros
    def elim_zeros(self):
        self.nz_indices, self.data = hlp.elim_zeros(self.nz_indices, self.data)

    # Builds up the sparse matrix in coo format
    # The coo format saves an array of row/column indices and the corresponding entries
    # Shape is the shape the sparse matrix should have
    def build_sr_matrix(self, shape, transpose=False):
        if prod(shape) != self.size:
            raise Exception("Tried to build sparse matrix with invalid shape")

        # Perform potential swaps
        self.apply_swaps()

        # Retrieve coordinates
        row, col = hlp.get_coordinates(self.nz_indices, shape)

        # Create sparse matrix
        if not transpose:
            res = coo_array((self.data, (row, col)), shape=shape)
        else:
            res = coo_array((self.data, (col, row)), shape=shape[::-1])

        # This attribute seems to be False which triggers a sorting algorithm,
        # which is in many cases among the most resource costly operations during
        # simulation. Setting this to True had no influence on the results tested so
        # far.
        res.has_canonical_format = True

        return res

    # This method allows moving an index to the first position without
    # changing the order of the remaining indices
    def move_index_to_front(self, index):
        # if the index is already at the front, we don't need to do anything
        if index == 0:
            return

        # The logic for creating the ne self.index_bit_permutation attribute is similar
        # as in swapaxes

        bit_shape = [int(np.log2(s)) for s in self.shape]

        new_shape = list(self.shape)
        new_shape.insert(0, new_shape.pop(index))

        self.shape = tuple(new_shape)

        index_range_start = sum(bit_shape[:index])
        index_range_end = index_range_start + bit_shape[index]

        init_perm = self.index_bit_permutation[:index_range_start]
        index_perm = self.index_bit_permutation[index_range_start:index_range_end]
        final_perm = self.index_bit_permutation[index_range_end:]
        self.index_bit_permutation = index_perm + init_perm + final_perm

    # This function performs tensor contraction along the specified axes.
    # Axes can be contracted if they have the same dimension
    # Imagine we have two tensors with shapes (2,4,64,8,16) and (8,32,4,2)
    # then we could for instance contract the following axes
    # 0 with 3
    # 1 with 2
    # 3 with 0
    # From this we will then get a tensor, which has the shape of the
    # first concatenated with the second but with the contraction indices removed
    # ie. (64,16,32)
    # The idea of this function is to move all the contraction indices to the front
    # reshape to matrices, transpose the first matrix, and then call some matrix
    # multiplication algorithm. In the case of the example the shapes after
    # moving to the front look like
    # (2,4,8,64,16) and (2,4,8,32)
    # After reshaping to matrices the shapes look like
    # (2*4*8, 64*16) and (2*4*8, 32)
    # Transpose the first matrix
    # (64*16, 2*4*8) and (2*4*8, 32)
    # Call matrix multiplication to get a matrix of shape
    # (64*16,32)
    # Finally, reshape the result
    # (64,16,32)
    def contract(self, other, axes_self, axes_other):
        # Check if the contraction size of both tensors is allowed
        contraction_size = prod([self.shape[i] for i in axes_self])
        contraction_size_other = prod([other.shape[i] for i in axes_other])

        if contraction_size != contraction_size_other:
            raise Exception(
                "Tried to contract tensor with differently sized contraction indices"
            )

        # Copy axes list in order to prevent modification
        axes_self = list(axes_self)
        axes_other = list(axes_other)

        # Move axes to the front
        for i in range(len(axes_self)):
            self.move_index_to_front(axes_self[i])

            # Note that we need to increase the index of the axes
            # greater than the axes we just moved, because
            # the insertion at the front moved the all the indices
            # before axes_self[i]
            for j in range(i + 1, len(axes_self)):
                if axes_self[i] > axes_self[j]:
                    axes_self[j] += 1

        # Perform the same for the other tensor
        for i in range(len(axes_other)):
            other.move_index_to_front(axes_other[i])

            for j in range(i + 1, len(axes_other)):
                if axes_other[i] > axes_other[j]:
                    axes_other[j] += 1

        # Determine new shape (now that the contraction axes are at the front, we can
        # simply slice the shapes)
        res_shape = list(self.shape[len(axes_self) :]) + list(
            other.shape[len(axes_other) :]
        )

        # In order to feature multiprocessing, we create an "empty" result array, which
        # will be returned after the calculation has been started. If the content is
        # needed before the calculation finished, the catch_up method will make sure the
        # calculation finishes before further processing
        res = SparseBiArray(
            (np.zeros(1, dtype=np.int64), np.zeros(1, dtype=self.data.dtype)),
            shape=res_shape,
            contraction_counter=self.contraction_counter + other.contraction_counter,
        )

        # Define multi processing wrapper
        def mp_wrapper():
            # Build up sparse matrices
            sr_matrix_self = self.build_sr_matrix(
                (contraction_size, self.size // contraction_size), transpose=True
            )
            sr_matrix_other = other.build_sr_matrix(
                (contraction_size, other.size // contraction_size)
            )

            # Perform sparse matrix multiplication

            res_sr_matrix = sparse_matrix_mult(sr_matrix_self, sr_matrix_other).tocoo()

            # Acquire flattened coordinates from the helper function
            res.nz_indices = hlp.gen_flat_coords(
                res_sr_matrix.col, res_sr_matrix.row, res_sr_matrix.shape
            )

            # Set the data
            res.data = res_sr_matrix.data

            # res.nz_indices, res.data = hlp.elim_zeros(res.nz_indices, res.data)

            # Set the sparsity
            res.sparsity = len(res.nz_indices) / res.size

            # Perform consolidation operations
            if False:
                # if res.contraction_counter > self.contraction_counter_threshold:
                res.sum_duplicates()
                res.eliminate_zeros()
                res.contraction_counter = 0

        # Start the thead
        if res.size > multithreading_threshold:
            res.thread = threading.Thread(target=mp_wrapper)
            res.thread.start()
        else:
            mp_wrapper()

        return res

    # This method returns self as a numpy array
    def to_array(self):
        # Apply all swapping operations
        self.apply_swaps()

        # Construct the flat array
        res = hlp.construct_flat_array(self.nz_indices, self.data, self.size)

        # Reshape
        return res.reshape(self.shape)

    def ravel(self):
        return self.reshape(self.size)

    def __repr__(self):
        return str(self.to_array())

    # This function allows to split the array into it's lower and upper half
    # i.e. if this SparseBiArray represents the array [0,1,2,3,4,5,6,7]
    # we return [0,1,2,3] and [4,5,6,7] as SparseBiArrays
    # This feature is important when measuring QuantumState as the norms of the
    # lower and upper half of the array determine the probability of measuring a 1 or 0
    def split(self):
        # Check if self is in a suitable shape (multi-dimensional arrays can not be
        # split as intended)
        if len(self.shape) != 1:
            raise Exception("Tried to split multi dimensional sparse array")

        # Applya swaps
        self.apply_swaps()

        # Find the boolean mask of indizes that are less than half of the size
        mask = self.nz_indices < self.size // 2

        # Acquire the lower half nz_indices
        lower_half_nz_indices = self.nz_indices[mask].copy()
        # Acquire the lower half data
        lower_half_data = self.data[mask].copy()

        # Create lower half result
        lower_half = SparseBiArray(
            (lower_half_nz_indices, lower_half_data), shape=(self.shape[0] // 2,)
        )

        # Perform the same process for the upper half but with inverted mask
        mask = np.invert(mask)
        upper_half_nz_indices = self.nz_indices[mask] - self.size // 2
        upper_half_data = self.data[mask].copy()

        upper_half = SparseBiArray(
            (upper_half_nz_indices, upper_half_data), shape=(self.shape[0] // 2,)
        )

        # Return result
        return lower_half, upper_half

    def multi_split(self, indices):
        sprs = self.build_sr_matrix(
            [2 ** (len(indices)), self.size // 2 ** len(indices)]
        ).tocsr()

        col_indices = sprs.indices
        row_ptr = sprs.indptr
        data = sprs.data

        new_bi_arrays = []
        for r in range(len(row_ptr) - 1):
            if row_ptr[r] == row_ptr[r + 1]:
                new_bi_array = None
            else:
                new_bi_array = SparseBiArray(
                    (
                        col_indices[row_ptr[r] : row_ptr[r + 1]].astype(np.int64),
                        data[row_ptr[r] : row_ptr[r + 1]],
                    ),
                    shape=(self.size // 2 ** len(indices),),
                )

            new_bi_arrays.append(new_bi_array)

        return new_bi_arrays

    # Calculate the squared norm, ie. the sesquilinear scalar product of self with
    # itself
    def squared_norm(self):
        return np.abs(np.vdot(self.data, self.data))

    # Return self as a DenseBiArray
    def to_dense(self):
        res = DenseBiArray(np.zeros(1), sparsity=self.sparsity, shape=self.shape)

        # We perform the same technique to enable multithreading as in contract,
        # ie. return an unfinished result that is waited on if required
        def mp_wrapper():
            res.data = self.to_array()

        # Start the thead
        if self.size > multithreading_threshold:
            res.thread = threading.Thread(target=mp_wrapper)
            res.thread.start()
        else:
            mp_wrapper()

        return res

    def to_sparse(self):
        return self


# This class serves mainly as an interface to numpy arrays using the same methods as the
# SparseBiArray class this way the algorithm using these classes doesn't need to care
# about wether it's treating a sparse array or a dense array.
# Furthermore, some multithreading techniques from the SparseBiArray class are deployed.
class DenseBiArray(BiArray):
    # The constructor can only be called with numpy arrays
    def __init__(self, array, sparsity=None, shape=None):
        if not isinstance(array, np.ndarray):
            raise Exception(
                "Tried to initialize DenseBiArray with type " + str(type(array))
            )

        # Save data
        self.data = array

        # Set shape
        if shape:
            self.shape = shape
        else:
            self.shape = array.shape

        # Determine size
        self.size = prod(self.shape)

        # If not given, estimate sparsity. In many situations the sparsity
        # of a contraction result can be estimated (and then given to the constructor),
        # so it is helpfull to track it
        if sparsity is None:
            self.sparsity = np.count_nonzero(array) / array.size
        else:
            # self.sparsity = np.count_nonzero(array)/array.size
            self.sparsity = sparsity

        # This attribute serves a similar purpose as in the SparseBiArray class
        self.index_bit_permutation = list(range(int(np.log2(self.size))))

        self.dtype = self.data.dtype

    # Copy method
    def copy(self):
        self.apply_swaps()
        return DenseBiArray(self.data.copy(), sparsity=self.sparsity)

    # Catch up method (same purpose as in SparseBiArray)
    def catch_up(self):
        if hasattr(self, "thread"):
            start_time = time.time()
            self.thread.join()

            duration = time.time() - start_time
            if duration > 0.001:
                # print("Waited ", duration)
                pass
            del self.thread

    # The shape of the DenseBiArray is tracked only as a tuple.
    # Once the data is required for a contraction, the actual data will be reshaped
    def reshape(self, shape):
        if isinstance(shape, int):
            shape = (shape,)

        if prod(shape) != prod(self.shape):
            raise Exception("Tried to reshape with invalid shape")

        self.shape = tuple(shape)
        return self

    # This method works similarly as it's equivalent in SparseBiArray
    def swapaxes(self, i, j):
        if i == j:
            return
        if i > j:
            i, j = j, i

        bit_shape = np.log2(self.shape).astype(np.int64)

        i_interval_start = sum(bit_shape[:i])
        i_interval_end = i_interval_start + bit_shape[i]

        j_interval_start = sum(bit_shape[:j])
        j_interval_end = j_interval_start + bit_shape[j]

        init_interval = self.index_bit_permutation[:i_interval_start]
        i_interval = self.index_bit_permutation[i_interval_start:i_interval_end]

        middle_interval = self.index_bit_permutation[i_interval_end:j_interval_start]

        j_interval = self.index_bit_permutation[j_interval_start:j_interval_end]

        final_interval = self.index_bit_permutation[j_interval_end:]

        self.index_bit_permutation = (
            init_interval + j_interval + middle_interval + i_interval + final_interval
        )

        new_shape = list(self.shape)

        new_shape[i], new_shape[j] = new_shape[j], new_shape[i]

        self.shape = tuple(new_shape)

        return self

    # This method works similarly as it's equivalent in SparseBiArray
    def move_index_to_front(self, index):
        if index == 0:
            return

        bit_shape = [int(np.log2(s)) for s in self.shape]

        new_shape = list(self.shape)
        new_shape.insert(0, new_shape.pop(index))

        self.shape = tuple(new_shape)

        index_range_start = sum(bit_shape[:index])
        index_range_end = index_range_start + bit_shape[index]

        init_perm = self.index_bit_permutation[:index_range_start]
        index_perm = self.index_bit_permutation[index_range_start:index_range_end]
        final_perm = self.index_bit_permutation[index_range_end:]
        self.index_bit_permutation = index_perm + init_perm + final_perm

    # This method works similarly as it's equivalent in SparseBiArray
    def apply_swaps(self):
        # Wait for any computation processes to finish
        self.catch_up()

        # Check if axes need to be swapped
        if self.index_bit_permutation == list(range(int(np.log2(self.size)))):
            # Check if data needs to be reshaped
            if self.data.shape != self.shape:
                self.data = self.data.reshape(self.shape)

            return

        # The plan is now to reshape into "tensor format" ie. log2(self.size) axes of
        # imension two and permute the axes

        # Numpy however does not allow arrays with more than 32 axes, so it would not be
        # possible to manipulate arrays describing more than 32 qubits.
        # In order to reduce the number of axes that need to be permuted,
        # we identify intervals of indices that do not need permutation on them
        # and group them into one axis

        # Get the permutation of the axes
        bi_index_perm = [
            self.index_bit_permutation.index(i) for i in range(int(np.log2(self.size)))
        ]

        if False:
            # result = np.empty(self.data.size, dtype = self.data.dtype)

            index_perm = hlp.invert_permutation(np.array(bi_index_perm))[::-1]
            index_perm = np.abs(len(index_perm) - 1 - index_perm)

            # flattened index array
            f_index_array = np.empty(2 ** len(index_perm), dtype=np.int64)

            f_index_array[0] = 0
            # result[0] = self.data.ravel()[0]

            # Unfortunately, this is not faster (maybe if implemented in C?)
            self.data = hlp.bi_array_moveaxis(
                self.data.ravel(), index_perm, f_index_array
            )
            # temp = hlp.bi_array_moveaxis(self.data.ravel(), index_perm, f_index_array)

            # permutated_indices = hlp.bi_array_moveaxis_jitted(index_perm, f_index_array)
            # self.data = self.data.ravel()[temp.ravel()]

            self.data.shape = self.shape

        else:
            perm = list(bi_index_perm)
            # Initialize the shape
            perm_shape = len(self.index_bit_permutation) * [2]

            i = 1
            while i < len(perm):
                # If this condition is met, both indices form an interval,
                # that does not need further permutation
                if perm[i] - 1 == perm[i - 1]:
                    # Remove the index from the perm and perm_shape list
                    perm.pop(i)
                    perm_shape.pop(i)

                    # Update the shape
                    perm_shape[i - 1] *= 2

                    # Update the permutation indices
                    for j in range(len(perm)):
                        if perm[j] > perm[i - 1]:
                            perm[j] -= 1
                else:
                    i += 1

            # Reshape data to the required shape
            self.data = self.data.reshape(perm_shape)

            # Move the axes on the data
            self.data = np.moveaxis(self.data, list(range(len(perm))), perm)

            # Reshape the data
            self.data = self.data.reshape(self.shape)

        # Set the index_bit_permutation to identity
        self.index_bit_permutation = list(range(int(np.log2(self.size))))

    # Raveling method
    def ravel(self):
        self.reshape(self.size)
        self.apply_swaps()
        return self

    # This function works similar as it's equivalent in SparseBiArray
    # i.e. identify contraction axes, move to the front, transpose one matrix
    # and then apply matrix multiplication
    def contract(self, other, axes_self, axes_other):
        # Check if the contraction size of both tensors is allowed
        contraction_size = prod([self.shape[i] for i in axes_self])
        contraction_size_other = prod([other.shape[i] for i in axes_other])

        if contraction_size != contraction_size_other:
            raise Exception(
                "Tried to contract tensor with differently sized contraction indices"
            )

        # Copy axes list in order to prevent modification
        axes_self = list(axes_self)
        axes_other = list(axes_other)

        # Move axes to the front
        for i in range(len(axes_self)):
            self.move_index_to_front(axes_self[i])

            # Note that we need to increase the index of the axes
            # greater than the axes we just moved, because
            # the insertion at the front moved the all the indices
            # before axes_self[i]
            for j in range(i + 1, len(axes_self)):
                if axes_self[i] > axes_self[j]:
                    axes_self[j] += 1

        # Perform the same for the other tensor
        for i in range(len(axes_other)):
            other.move_index_to_front(axes_other[i])

            for j in range(i + 1, len(axes_other)):
                if axes_other[i] > axes_other[j]:
                    axes_other[j] += 1

        # Determine new shape (now that the contraction axes are at the front, we can
        # simply slice the shapes).
        res_shape = list(self.shape[len(axes_self) :]) + list(
            other.shape[len(axes_other) :]
        )

        # The sparsity can be estimated using statistical methods:
        # In a regular tensor contraction, any entry of the resulting tensor, is the sum
        # of n = contraction_size products of entries of the two source tensors.
        # Each of these products is non-zero with probability
        # p_nz = self.sparsity*other.sparsity

        # p_nz = self.sparsity*other.sparsity

        # The probabilty for the product being zero is therefore
        # p_z = 1 - self.sparsity*other.sparsity

        # p_z = 1 - self.sparsity*other.sparsity

        # Therefore, we perform n = contraction_size random experiments with probability
        # p_z. The result entry is zero if all products are zero, so

        # p_res_z = p_z**contraction_size

        # To get the sparsity, we finally need the probability of the result entry being
        # non-zero.

        # res_sparsity = 1-p_res_z

        # We tested this formula and it works extremely well.

        res_sparsity = 1 - (1 - self.sparsity * other.sparsity) ** contraction_size

        # Prepare "unfinished" result
        res = DenseBiArray(
            np.array([self.data.ravel()[0] * other.data.ravel()[0]]),
            sparsity=res_sparsity,
            shape=res_shape,
        )

        # Prepare multithreading wrapper
        # Performs the same logic as it's SparseBiArray equivalent
        def mp_wrapper():
            original_shape_self = self.shape
            original_shape_other = other.shape

            self.reshape([contraction_size, self.size // contraction_size])
            other.reshape([contraction_size, other.size // contraction_size])

            self.swapaxes(0, 1)

            self.apply_swaps()
            other.apply_swaps()

            res.data = np.matmul(self.data, other.data)

            self.swapaxes(0, 1)
            self.reshape(original_shape_self)
            other.reshape(original_shape_other)

        if res.size > multithreading_threshold:
            # Start the wrapper
            res.thread = threading.Thread(target=mp_wrapper)
            res.thread.start()
        else:
            mp_wrapper()
        # Return the result
        return res

    # This method works similarly as it's equivalent in SparseBiArray
    def split(self):
        if len(self.shape) != 1:
            raise Exception("Tried to split multi rank BiArray")
        self.apply_swaps()

        return DenseBiArray(
            self.data[: self.size // 2].copy(), self.sparsity
        ), DenseBiArray(self.data[self.size // 2 :].copy(), sparsity=self.sparsity)

    def multi_split(self, indices):
        original_shape = tuple(self.shape)
        self.reshape([2 ** (len(indices)), self.size // 2 ** len(indices)])

        np_array = self.to_array()

        new_bi_arrays = []

        for r in range(2 ** len(indices)):
            new_bi_arrays.append(DenseBiArray(np_array[r, :]))

        self.reshape(original_shape)

        return new_bi_arrays

    # This method works similarly as it's equivalent in SparseBiArray
    def squared_norm(self):
        self.apply_swaps()
        return np.abs(np.vdot(self.data, self.data))

    # This method works similarly as it's equivalent in SparseBiArray
    def to_array(self):
        self.apply_swaps()
        return self.data

    # This method works similarly as it's equivalent in SparseBiArray
    def to_dense(self):
        return self

    # Generates the SparseBiArray version of self using multithreading
    def to_sparse(self):
        res = SparseBiArray(
            (np.zeros(1, dtype=np.int64), np.zeros(1)), shape=self.shape
        )

        # Prepare the wrapper
        def mp_wrapper():
            self.apply_swaps()

            # Ravel init array
            raveled_array = self.data.ravel()

            # Find non-zero indices
            res.nz_indices = np.nonzero(raveled_array)[0]

            # Store data
            res.data = raveled_array[res.nz_indices]

            res.nz_indices, res.data = hlp.elim_zeros(res.nz_indices, res.data)

            res.sparsity = len(res.nz_indices) / res.size

        # Start the
        if res.size > multithreading_threshold:
            res.thread = threading.Thread(target=mp_wrapper)
            res.thread.start()
        else:
            mp_wrapper()
        return res

    def __repr__(self):
        return str(self.to_array())


# This class is used to estimate the complexity of a contraction order
# The contract method does not actually execute a contraction but only logs
# the estimated amount of floating point operations
class DummyBiArray(BiArray):
    current_memory = np.zeros(1)
    max_memory = np.zeros(1)
    flops = np.zeros(1)

    def __init__(self, data=None, shape=None, sparsity=1):
        self.sparsity = sparsity
        if not isinstance(data, type(None)):
            self.shape = data.shape
            self.sparsity = np.count_nonzero(data) / data.size
        if not isinstance(shape, type(None)):
            self.shape = tuple(shape)

        self.size = prod(self.shape)

    def reshape(self, shape):
        self.shape = shape

    def swapaxes(self, axis_0, axis_1):
        self.shape = list(self.shape)
        self.shape[axis_0], self.shape[axis_1] = self.shape[axis_1], self.shape[axis_0]
        self.shape = tuple(self.shape)

    def contract(self, other, axes_self, axes_other):
        contraction_size = prod([self.shape[i] for i in axes_self])

        temp_shape_self = list(self.shape)
        temp_shape_other = list(other.shape)

        delete_multiple_element(temp_shape_self, axes_self)
        delete_multiple_element(temp_shape_other, axes_other)

        res_sparsity = 1 - (1 - self.sparsity * other.sparsity) ** contraction_size
        res_shape = temp_shape_self + temp_shape_other
        try:
            self.flops += prod(res_shape) * contraction_size
        except TypeError:
            self.flops += np.inf

        res = DummyBiArray(shape=res_shape, sparsity=res_sparsity)
        self.current_memory[0] += res.size - self.size - other.size

        if self.current_memory > self.max_memory:
            self.max_memory[0] = self.current_memory[0]

        return res

    def copy(self):
        return DummyBiArray(shape=self.shape)

    @classmethod
    def reset_stat_counter(self):
        self.current_memory = np.zeros(1)
        self.max_memory = np.zeros(1)
        self.flops = np.zeros(1)


# This function emulates the behavior of np.tensordot
def tensordot(a, b, axes, contract_sparsity_threshold=0.01):
    # Turn axes object into a list
    axes = list(axes)
    if isinstance(axes[0], int):
        axes[0] = [axes[0]]
    if isinstance(axes[1], int):
        axes[1] = [axes[1]]

    # Treat the case of a dummy contraction
    if isinstance(a, DummyBiArray) and isinstance(b, DummyBiArray):
        return a.contract(b, axes[0], axes[1])

    # Treat the case of a numpy contraction
    if isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
        return np.tensordot(a, b, axes)

    if a.data.dtype == np.dtype("O") or b.data.dtype == np.dtype("O"):
        a = a.to_dense()
        b = b.to_dense()

        return a.contract(b, axes[0], axes[1])

    if len(axes[0]) != 0:
        contraction_size = prod([a.shape[i] for i in axes[0]])

        # Estimate resulting sparsity (for more information on this estimate check the
        # contract method of DenseBiArray)
        res_sparsity = 1 - (1 - a.sparsity * b.sparsity) ** contraction_size

        res_size = a.size * b.size / contraction_size
    else:
        res_sparsity = a.sparsity * b.sparsity

        res_size = a.size * b.size

    if (
        res_sparsity > contract_sparsity_threshold or a.size * b.size < 2**12
    ) and not res_size > 2**32:
        #     if ((isinstance(a, SparseBiArray) and isinstance(b, DenseBiArray)) or
        #     (isinstance(b, SparseBiArray) and isinstance(a, DenseBiArray))):

        #         return mixed_contractor(a, b, axes)

        a = a.to_dense()
        b = b.to_dense()

    else:
        a = a.to_sparse()
        b = b.to_sparse()

    # Treat the case of a "contractionless" tensor product
    if len(axes[0]) == 0:
        # We utilize the multithreading pattern presented in the previous contraction
        # methods.
        if isinstance(a, SparseBiArray) and isinstance(b, SparseBiArray):
            # Prepare result arrays
            res = SparseBiArray(
                (np.zeros(1, dtype=np.int64), np.ones(1)),
                shape=list(a.shape) + list(b.shape),
            )

            # Prepare wrapper
            def mp_wrapper():
                # Apply any potential swaps on the source arrays
                a.apply_swaps()
                b.apply_swaps()

                # Imagine we have two arrays with flattened nz_indices [1,2,3], [2,3,6]
                # and sizes 16, 8  respectively
                # the nz_indices of the result can be ordered in a matrix
                # [1,2,3] + 2*16
                # [1,2,3] + 3*16
                # [1,2,3] + 6*16
                # (where the plus is executed on all entries)

                # This idea is encapsulated in the following command
                nz_indices = np.tensordot(
                    b.size * a.nz_indices,
                    np.ones(len(b.data), dtype=np.int64),
                    ((), ()),
                ) + np.tensordot(
                    np.ones(len(a.data), dtype=np.int64), b.nz_indices, ((), ())
                )

                # The corresponding data can be calculated as
                data = np.tensordot(a.data.ravel(), b.data.ravel(), ((), ())).ravel()

                # Set nz_indices and data
                res.nz_indices = nz_indices.ravel()

                res.data = data

            if res.size > multithreading_threshold:
                res.thread = threading.Thread(target=mp_wrapper)
                res.thread.start()
            else:
                mp_wrapper()

            return res

        else:
            # For the DenseBiArray, we can simply use the numpy tensordot
            a.catch_up()
            b.catch_up()

            res = DenseBiArray(
                np.array([a.data.ravel()[0] * b.data.ravel()[0]]),
                sparsity=a.sparsity * b.sparsity,
                shape=list(a.shape) + list(b.shape),
            )

            def mp_wrapper():
                a.apply_swaps()
                b.apply_swaps()

                res.data = np.tensordot(a.to_array(), b.to_array(), axes=((), ()))

            if res.size > multithreading_threshold:
                res.thread = threading.Thread(target=mp_wrapper)
                res.thread.start()
            else:
                mp_wrapper()

            return res

    return a.contract(b, axes[0], axes[1])


def mixed_contractor(a, b, axes):
    if not (
        (isinstance(a, SparseBiArray) and isinstance(b, DenseBiArray))
        or (isinstance(b, SparseBiArray) and isinstance(a, DenseBiArray))
    ):
        raise Exception("Tried to call mixed contraction on non-mixed input")

    # Copy axes list in order to prevent modification
    axes_a = list(axes[0])
    axes_b = list(axes[1])

    # Check if the contraction size of both tensors is allowed
    contraction_size = prod([a.shape[i] for i in axes_a])
    contraction_size_b = prod([b.shape[i] for i in axes_b])

    if contraction_size != contraction_size_b:
        raise Exception(
            "Tried to contract tensor with differently sized contraction indices"
        )

    # Move axes to the front
    for i in range(len(axes_a)):
        a.move_index_to_front(axes_a[i])

        # Note that we need to increase the index of the axes
        # greater than the axes we just moved, because
        # the insertion at the front moved the all the indices
        # before axes_a[i]
        for j in range(i + 1, len(axes_a)):
            if axes_a[i] > axes_a[j]:
                axes_a[j] += 1

    # Perform the same for the b tensor
    for i in range(len(axes_b)):
        b.move_index_to_front(axes_b[i])

        for j in range(i + 1, len(axes_b)):
            if axes_b[i] > axes_b[j]:
                axes_b[j] += 1

    # Determine new shape (now that the contraction axes are at the front, we can simply
    # slice the shapes)
    res_shape = list(a.shape[len(axes_a) :]) + list(b.shape[len(axes_b) :])

    res_sparsity = 1 - (1 - a.sparsity * b.sparsity) ** contraction_size

    # Prepare "unfinished" result
    res = DenseBiArray(
        np.empty(1, dtype=a.data.dtype), sparsity=res_sparsity, shape=res_shape
    )

    # Define multi processing wrapper
    def mp_wrapper():
        # Build up sparse matrices
        if isinstance(a, SparseBiArray):
            mult_matrix_a = a.build_sr_matrix(
                (contraction_size, a.size // contraction_size), transpose=True
            )
            original_shape_b = b.shape
            b.reshape([contraction_size, b.size // contraction_size])
            b.apply_swaps()
            mult_matrix_b = b.data

        else:
            original_shape_a = a.shape
            a.reshape([contraction_size, a.size // contraction_size])
            a.swapaxes(0, 1)
            a.apply_swaps()
            mult_matrix_a = a.data
            mult_matrix_b = b.build_sr_matrix(
                (contraction_size, b.size // contraction_size)
            )

        # Set the data
        res.data = sparse_matrix_mult(mult_matrix_a, mult_matrix_b)

    # Start the thead
    if res.size > multithreading_threshold:
        res.thread = threading.Thread(target=mp_wrapper)
        res.thread.start()
    else:
        mp_wrapper()

    return res


def delete_multiple_element(list_object, indices):
    indices = sorted(indices, reverse=True)
    for idx in indices:
        if idx < len(list_object):
            list_object.pop(idx)
